#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import numpy as np
import astropy.units as u
from astropy import log
from pint.utils import weighted_mean
import pint.residuals as Resid
import pint.models.parameter
from pint.models import get_model
import os
import time
from subprocess import check_output, check_call, CalledProcessError
import glob
# Import some software so we have appropriate versions
import pint
import astropy
# import enterprise_extensions as e_e # NOTE - enterprise_extensions has no attribute __version__
from pint_pal.ftester import get_fblist, param_check, summarize_Ftest
import scipy.stats
import copy

ALPHA = 0.0027

def whiten_resids(fitter, restype = 'postfit'):
    """
    Function to whiten residuals. If no reddened residuals, input will be returned.

    Inputs:
    ---------
    fitter [object, dictionary]: PINT fitter class or dictionary output from ecorr_average() function.
    restype ['string']: Type of residuals, pre or post fit, to plot from fitter object. Options are:
        'prefit' - plot the prefit residuals.
        'postfit' - plot the postfit residuals (default)

    Returns:
    ---------
    wres [array]: Array of whitened timing residuals.
    """
    # Check if input is the epoch averaged dictionary, should only be used if epoch averaged NB TOAs
    if type(fitter) is dict:
        rs = fitter['time_resids']
        noise_rs = fitter['noise_resids']
        # Now check if red noise residuals
        if "pl_red_noise" and "pl_DM_noise" in noise_rs:
            wres = rs - noise_rs['pl_red_noise'] - noise_rs['pl_DM_noise']
        elif "pl_red_noise" in noise_rs:
            wres = rs - noise_rs['pl_red_noise']
        elif "pl_DM_noise" in noise_rs:
            wres = rs - noise_rs['pl_DM_noise']
        else:
            log.warning("No red noise, residuals already white. Returning input residuals...")
            wres = rs
    # if not assume it's a PINT fitter class object
    else:
        # Check if WB or NB
        if fitter.is_wideband:
            if restype == 'postfit':
                time_resids = fitter.resids.residual_objs['toa'].time_resids
                noise_resids = fitter.resids.noise_resids
            else:
                time_resids = fitter.resids_init.residual_objs['toa'].time_resids
                noise_resids = fitter.resids_init.noise_resids
        else:
            if restype == 'postfit':
                time_resids = fitter.resids.time_resids
                noise_resids = fitter.resids.noise_resids
            elif restype == 'prefit':
                time_resids = fitter.resids_init.time_resids
                noise_resids = fitter.resids_init.noise_resids
            else:
                raise ValueError("Unrecognized residual type: %s. Please choose from 'prefit' or 'postfit'."%(restype))
            # Get number of residuals
        num_res = len(time_resids)
        # Check that the key is in the dictionary
        if "pl_red_noise" and "pl_DM_noise" in noise_resids:
            wres = time_resids - noise_resids['pl_red_noise'][:num_res] - noise_resids['pl_DM_noise'][:num_res]
        elif "pl_red_noise" in noise_resids:
            wres = time_resids - noise_resids['pl_red_noise'][:num_res]
        elif "pl_DM_noise" in noise_resids:
            wres = time_resids - noise_resids['pl_DM_noise'][:num_res]
        else:
            log.warning("No red noise, residuals already white. Returning input residuals...")
            wres = time_resids
    return wres

def rms_by_backend(resids, errors, rcvr_backends, dm = False):
    """
    Function to take a fitter, list of residuals errors, and backends and compute the rms and weighted rms
    for either time residuals or DM residuals if a wideband residuals.

    Inputs:
    ----------
    resids [list]: List of residuals.
    errors [list]: List of residual errors.
    rcvr_backends [list]: List of backends.
    dm [boolean]: If True, will do computation with DM residuals [defaut: False].
    
    Returns:
    ----------
    rs_dict [dictionary]: Dictionary of rms and wieghted rms residuals for each backend-reciever combination.
    """
    # Define output dictionary
    rs_dict = {}
    # Get RMS of all residuals
    # Now loop through and compute on a per receiver-backend status
    RCVR_BCKNDS = np.sort(list(set(rcvr_backends)))
    if dm:
        avg_RMS_ALL = np.std(resids)
        # Get the weighted rms of averaged residuals
        weights = 1.0 / (errors ** 2)
        wmean, werr, wsdev = weighted_mean(resids, weights, sdev=True)
        avg_WRMS_ALL = wsdev
        # Add to dictionary
        rs_dict['All'] = {'rms':avg_RMS_ALL, 'wrms':avg_WRMS_ALL}
        for r_b in RCVR_BCKNDS:
            # Get indices of receiver-backend toas
            inds = np.where(rcvr_backends==r_b)[0]
            # Select them
            rms = np.std(resids[inds])
            weights = 1.0 / (errors ** 2)
            wmean, werr, wsdev = weighted_mean(resids[inds], weights[inds], sdev=True)
            wrms = wsdev
            rs_dict[r_b] = {'rms':rms, 'wrms':wrms}
    else:
        avg_RMS_ALL = np.std(resids).to(u.us)
        # Get the weighted rms of averaged residuals
        weights = 1.0 / (errors.to(u.s) ** 2)
        wmean, werr, wsdev = weighted_mean(resids, weights, sdev=True)
        avg_WRMS_ALL = wsdev.to(u.us)
        # Add to dictionary
        rs_dict['All'] = {'rms':avg_RMS_ALL, 'wrms':avg_WRMS_ALL}
        for r_b in RCVR_BCKNDS:
            # Get indices of receiver-backend toas
            inds = np.where(rcvr_backends==r_b)[0]
            # Select them
            rms = np.std(resids[inds].to(u.us))
            weights = 1.0 / (errors.to(u.s) ** 2)
            wmean, werr, wsdev = weighted_mean(resids[inds], weights[inds], sdev=True)
            wrms = wsdev.to(u.us)
            rs_dict[r_b] = {'rms':rms, 'wrms':wrms}
    # return the dictionary
    return rs_dict


def resid_stats(fitter, epoch_avg = False, whitened = False, dm_stats = False, print_pretty = False):
    """
    Function to get statistics for the residuals. This includes the RMS and WRMS for all residuals, as well as
    per-backend. Option for epoch averaged or not epoch averaged. If dm_stats are also returned, then there
    will be a second output dictionary for the DM stats by receiver-backend combo.

    Inputs:
    ----------
    fitter [object]: PINT fitter class object, post-fit.
    epoch_avg [boolean]: If True, will output stats for epoch averaged residulas, else will be for 
        non-epoch averaged residuals [default: False].
    whitened [boolean]: If True, will output stats for whitened residulas, else will be for 
        non-whitened residuals [default: False].
    dm_stats [boolean]: If True, will also output the stats for the DM residuals for wideband fitters.
        Note this will return an additional dictionary, dm_stats [default: False].
    print_pretty [boolean]: If True, will nicely print the W/RMS per receiver-backend combo [default: False].

    Returns:
    ----------
    rs_dict [nested dictionary]:

        First set of keys are Receiver-Backend combos, e.g. L-wide_ASP, S-wide_PUPPI. For the W/RMS of all residuals,
        the key is 'All'.

        Within each Receiver-Backend combo the flags are:
            rms [astropy quantity: rms of the residuals for the receiver-backend combo in microseconds.
            wrms [astropy quantity: weighted rms of the residuals for the receiver-backend combo in microseconds.
     
    dm_dict [nested dictionary]: Same as rs_dict but for DM residuals with unit of pc cm^-3.
    """
    # Check if fitter is WB or not
    if fitter.is_wideband:
        resids = fitter.resids.residual_objs['toa']
        dm_resids = fitter.resids.residual_objs['dm']
        NB = False
        if epoch_avg:
            log.warning("Warning, cannot epoch average wideband residuals, will skip epoch averaging.")
    else:
        resids = fitter.resids
        NB = True

    # get rcvr backend combos for averaged residuals
    rcvr_bcknds = np.array(resids.toas.get_flag_value('f')[0])

    # Compute epoch averaged stats
    if epoch_avg and NB:
        avg = fitter.resids.ecorr_average(use_noise_model=True)
        avg_rcvr_bcknds = []
        for iis in avg['indices']:
            avg_rcvr_bcknds.append(rcvr_bcknds[iis[0]])
        avg_rcvr_bcknds = np.array(avg_rcvr_bcknds)
        # compute averaged, whitened
        if whitened:
            wres_avg = whiten_resids(avg)
            rs_dict = rms_by_backend(wres_avg.to(u.us), avg['errors'], avg_rcvr_bcknds)
        # compute averaged
        else:
            rs_dict = rms_by_backend(avg['time_resids'], avg['errors'], avg_rcvr_bcknds)

    # Compute whitened
    elif whitened:
        wres = whiten_resids(fitter)
        #rs_dict = rms_by_backend(wres.to(u.us), fitter.toas.get_errors(), rcvr_bcknds)
        rs_dict = rms_by_backend(wres.to(u.us), resids.get_data_error(), rcvr_bcknds)

    # If not averaged or whitened, compute with functions that already exist
    if not epoch_avg and not whitened:
        # Define dictionary for return values
        rs_dict = {}
        # Get total RMS and WRMS
        RMS_ALL = resids.time_resids.std().to(u.us) # astropy quantity
        WRMS_ALL = resids.rms_weighted() # astropy quantity, units are us

        # Now split up by backend
        RCVR_BCKNDS = np.sort(list(set(rcvr_bcknds)))
        # Turn into dictionary to return
        rs_dict['All'] = {'rms':RMS_ALL, 'wrms':WRMS_ALL}
        for r_b in RCVR_BCKNDS:
            # Get indices of receiver-backend toas
            inds = np.where(rcvr_bcknds==r_b)[0]
            # Select them
            fitter.toas.select(inds)
            # Create new residual object
            r = Resid.Residuals(fitter.toas, fitter.model)
            # Get new RMS, WRMS
            rs_dict[r_b] = {'rms':r.time_resids.std().to(u.us), 'wrms':r.rms_weighted()}
            # Unselect them
            fitter.toas.unselect()

    # print output if desired
    if print_pretty:
        rs_keys = rs_dict.keys()
        for k in rs_keys:
            l = "# WRMS(%s) = %.3f %s" %(k, rs_dict[k]['wrms'].value, rs_dict[k]['wrms'].unit)
            print(l)
            l = "#  RMS(%s) = %.3f %s" %(k, rs_dict[k]['rms'].value, rs_dict[k]['rms'].unit)
            print(l)

    # Check if dm stats are desired
    if dm_stats:
        if not NB:
            dm_dict = rms_by_backend(dm_resids.resids, dm_resids.get_data_error(), rcvr_bcknds, dm = True)
            if print_pretty:
                print()
                dm_keys = dm_dict.keys()
                for k in rs_keys:
                    l = "# WRMS(%s) = %.6f %s" %(k, dm_dict[k]['wrms'].value, dm_dict[k]['wrms'].unit)
                    print(l)
                    l = "#  RMS(%s) = %.6f %s" %(k, dm_dict[k]['rms'].value, dm_dict[k]['rms'].unit)
                    print(l)
        else:
            log.warning("Cannot compute DM Stats, not Wideband timing data.")

    # Return the dictionary
    if dm_stats and not NB:
        return rs_dict, dm_dict
    else:
        return rs_dict



# Define helper functions
def year(mjd):
    """
    Calculate the year from an MJD.
    
    Inputs:
    ---------
    mjd [float]: MJD value.
    
    Returns:
    ---------
    year [float]: MJD value converted to a year.
    """
    return (mjd - 51544.0)/365.25 + 2000.0


def report_ptest(label, ftest_dict = None, alpha=ALPHA):
    """
    Nicely prints the results of F-tests in a human-readable format.
    
    Input:
    --------
    label [string]: Name of the parameter(s) that were added/removed for the F-test.
    ftest_dict [dictionary]: Dictionary of output values from the PINT `ftest()` function. If `None`, will
        print a line of NaNs for each reported value.
    alpha [float]: Value to compare for F-statistic significance. If the F-statistic is lower than alpha, 
        the timing model parameters are deemed statistically significant to the timing model.
    """
    # If F-test fails, print line of NaNs
    if ftest_dict == None:
        line = "%42s %7.3f %9.2f %5f %.3f" % (label, np.nan, np.nan, np.nan, np.nan)
    # Else print the computed values
    else:
        # Get values from input dictionary
        rms = ftest_dict['resid_wrms_test'].value # weighted root mean square of timing residuals
        chi2 = ftest_dict['chi2_test'] # chi-squared value of the fit of the F-tested model
        ndof = ftest_dict['dof_test'] # number of degrees of freedom in the F-tested model
        if "dm_resid_wrms_test" in ftest_dict.keys():
            dmrms = ftest_dict['dm_resid_wrms_test'].value # weighted root mean square of DM residuals
        else:
            dmrms = None
        Fstatistic=  ftest_dict['ft'] # F-statistic from the F-test comparison
        if Fstatistic is None:
            if dmrms != None:
                line = "%42s %7.3f %16.6f %9.2f %5d --" % (label, rms, dmrms, chi2, ndof)
            else:
                line = "%42s %7.3f %9.2f %5d --" % (label, rms, chi2, ndof)
        elif Fstatistic:
            if dmrms != None:
                line = "%42s %7.3f %16.6f %9.2f %5d %.3g" % (label, rms, dmrms, chi2, ndof, Fstatistic)
            else:
                line = "%42s %7.3f %9.2f %5d %.3g" % (label, rms, chi2, ndof, Fstatistic)
            if Fstatistic < alpha:
                line += " ***"
        else:
            if dmrms != None:
                line = "%42s %7.3f %16.6f %9.2f %5d xxx" % (label, rms, dmrms, chi2, ndof)
            else:
                line = "%42s %7.3f %9.2f %5d xxx" % (label, rms, chi2, ndof)
    return line


def get_Ftest_lines(Ftest_dict, fitter, alpha = ALPHA):
    """
    Function to get nicely formatted lines from F-test dictionary.

    Input:
    ----------
    Ftest_dict [dictionary]: Dictionary of F-test results output by the `run_Ftests()` function.
    fitter [object]: The PINT fitter object.
    
    Returns:
    ----------
    ftest_lines [list]: List of nicely formatted F-test results lines to be printed elsewhere.
    """
    ftest_lines = []
    cur_fd = [param for param in fitter.model.params if "FD" in param]
    for fk in Ftest_dict.keys():
        # Get the FB parameter lines
        if 'FB' in fk:
            # Get the value of fbmax, note, may need fixes somewhere
            try:
                fbmax = (int(max(Ftest_dict[fk].keys())[-1]))
            except:
                fbmax = (int(max(Ftest_dict[fk].keys())[-2]))
            fblist = get_fblist(fitter)
            fbused = (len(fblist)>0)
            fbp = [fblist[ifb] for ifb in sorted(fblist.keys())]  # sorted list of fb parameters
            ftest_lines.append("\nTesting FB parameters, present list: "+" ".join(fbp))
            ftest_lines.append("\nTesting removal of FB parameters:")
            for i in range(1,len(fblist)):
                p = [fbp[j] for j in range(i,len(fbp))]
                ffk = 'FB%s+'%i
                l = report_ptest(" ".join(p), Ftest_dict[fk][ffk], alpha = alpha)
                ftest_lines.append(l)
            ftest_lines.append("Testing addition of FB parameters:")
            for i in range(len(fblist),fbmax+1):
                p = ["FB%d" % (j) for j in range(len(fblist),i+1)]
                ffk = 'FB%s'%i
                l = report_ptest(" ".join(p), Ftest_dict[fk][ffk], alpha = alpha)
                ftest_lines.append(l)
        # Report the intial values        
        elif 'initial' in fk:
            l = report_ptest(fk, Ftest_dict[fk])
            ftest_lines.append(l)
        # Report any added F-tested parameters, including FD
        elif "Add" in fk:
            ftest_lines.append('Testing additional parameters:')
            for ffk in Ftest_dict[fk].keys():
                if ffk == 'Binary':
                    for fffk in Ftest_dict[fk][ffk].keys():
                        l = report_ptest(fffk, Ftest_dict[fk][ffk][fffk], alpha = alpha)
                        ftest_lines.append(l)
                elif 'FD' in ffk:
                    ftest_lines.append("\nTesting adding FD terms (%s enabled):" % (cur_fd))
                    for fffk in Ftest_dict[fk][ffk].keys():
                        l = report_ptest(fffk, Ftest_dict[fk][ffk][fffk], alpha = alpha)
                        ftest_lines.append(l)
                else:
                    l = report_ptest(ffk, Ftest_dict[fk][ffk], alpha = alpha)
                    ftest_lines.append(l)
        # Report any removed F-tested parameters, including FD 
        elif "Remove" in fk:
            ftest_lines.append('\nTesting removal of parameters:')
            for ffk in Ftest_dict[fk].keys():
                if ffk == 'Binary':
                    for fffk in Ftest_dict[fk][ffk].keys():
                        l = report_ptest(fffk, Ftest_dict[fk][ffk][fffk], alpha = alpha)
                        ftest_lines.append(l)
                elif 'FD' in ffk:
                    ftest_lines.append("\nTesting removing FD terms (%s enabled):" % (cur_fd))
                    for fffk in Ftest_dict[fk][ffk].keys():
                        l = report_ptest(fffk, Ftest_dict[fk][ffk][fffk], alpha = alpha)
                        ftest_lines.append(l)
                else:
                    l = report_ptest(ffk, Ftest_dict[fk][ffk], alpha = alpha)
                    ftest_lines.append(l)
                    
        elif fk == 'F':
            # Get current number of spin frequency derivatives
            current_freq_deriv = 1
            for i in range(2,21):
                p = "F%d" % i
                if p in fitter.model.params:
                    current_freq_deriv = i
            ftest_lines.append("Testing spin freq derivs (%s enabled):" % (current_freq_deriv))
            for ffk in Ftest_dict[fk].keys():
                l = report_ptest(ffk, Ftest_dict[fk][ffk], alpha = alpha)
                ftest_lines.append(l)

    return ftest_lines

def interesting_compare_lines(cm):
    header_done = False
    for c in cm:
        cs = c.split()
        if not cs:
            continue
        if not header_done:
            if cs[0].startswith("----"):
                header_done = True
            continue
        if cs[0] in {
                "PSR", # Right at the top of the page
                "UNITS", "INFO", "TIMEEPH", "T2CMETHOD", "CHI2", 
                "POSEPOCH", "PEPOCH", "DMEPOCH",  # Epochs are forced equal
                "TZRMJD", "TZRSITE", "TZRFRQ", # Don't care about absolute phase
                "SWM"}:
            continue
        if any(cs[0].startswith(p) for p in ["JUMP", "EFAC", "EQUAD", "ECORR", "DMEFAC", "DMEQUAD", "DMJUMP"]):
            # Are these interesting? No guaraantee that JUMP1 refers to the same jump
            continue
        try:
            if len(cs)==3 and float(cs[1])==float(cs[2])==0.0:
                continue
        except ValueError:
            # not floats
            pass
        yield c

def pdf_writer(fitter, 
               parfile, 
               rs_dict, 
               Ftest_dict=None, 
               dm_dict=None, 
               append=None, 
               previous_parfile=None, 
               fitter_noise=None,
               cuts_dict=None,
               no_corner=False):
    """Take output from timing notebook functions and write things out nicely in a summary pdf.

    Input
    -----
    fitter [pint.fitter.Fitter]: Fitter used for initial fit.
    parfile [string]: Name of parfile used to generate residuals.
    rs_dict [dictionary]: Dictionary of residual stats output by the `resid_stats()` function.
    Ftest_dict [dictionary]: Dictionary of F-test results output by the `run_Ftests()` function.
    dm_dict [dictionary]: Optional dictionary of DM residual stats output by the `resid_stats()` function for WB timing.
        Input is optional. if `None` will not write out the DM residual stats [default: None].
    append [string or None]: default is `None`, else should be a string to the path to the texfile to append output to.
    previous_parfile [string or None]: If provided, report a comparison with this par file (presumably from a previous release).
    fitter_noise [pint.fitter.Fitter]: Fitter that has had new noise parameters applied (if available).
    cuts_dict [dictionary]: optional dictionary specifying total number TOAs/cuts by tel, febe, flag value
    no_corner [boolean]: default is `False` to append corner plot, else will look for posterior plots
    """
    def verb(s):
        s = str(s).strip()
        if "\n" in s:
            return "\n".join(verb(p) for p in s.split("\n"))
        for c in "@|!%":
            if c not in s:
                return r'\verb' + c + s + c
        else:
            raise ValueError(f"String {s} contains all my known verbatim quoting characters")
    def alert(s):
        s = str(s)
        return r"\textbf{\color{red}" + s + "}"
    
    # Check if fitter is wideband or not
    if fitter.is_wideband:
        NB = False
        resids = fitter.resids.residual_objs['toa']
        dm_resids = fitter.resids.residual_objs['dm']
    else:
        NB = True
        resids = fitter.resids
    
    # Start the latex pdf text (from old finalize timing script)
    psr = fitter.model.PSR.value.replace('-','$-$')
    write_header = True
    if append != None:
        texfile = append
        if os.path.exists(texfile):
            write_header = False
        fsum = open(texfile,'a')
    else:
        if NB:
            texfile = fitter.model.PSR.value + '.summary.nb.tex'
        else:
            texfile = fitter.model.PSR.value + '.summary.wb.tex'
        fsum = open(texfile,'w')
    
    if write_header:
        fsum.write(r'\documentclass[11pt]{article}' + '\n')
        fsum.write(r'\usepackage[T1]{fontenc}' + '\n')
        fsum.write(r'\usepackage[utf8]{inputenc}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{D7}{$\times$}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{B9}{\textsuperscript{1}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{B2}{\textsuperscript{2}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{B3}{\textsuperscript{3}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{207B}{\textsuperscript{-}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2070}{\textsuperscript{0}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2071}{\textsuperscript{1}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2072}{\textsuperscript{2}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2073}{\textsuperscript{3}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2074}{\textsuperscript{4}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2075}{\textsuperscript{5}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2076}{\textsuperscript{6}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2077}{\textsuperscript{7}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2078}{\textsuperscript{8}}' + '\n')
        fsum.write(r'\DeclareUnicodeCharacter{2079}{\textsuperscript{9}}' + '\n')
        fsum.write(r'\usepackage{graphicx}' + '\n')
        fsum.write(r'\usepackage{xcolor}' + '\n')
        fsum.write(r'\addtolength{\hoffset}{-2.5cm}' + '\n')
        fsum.write(r'\addtolength{\textwidth}{5.0cm}' + '\n')
        fsum.write(r'\addtolength{\voffset}{-2.5cm}' + '\n')
        fsum.write(r'\addtolength{\textheight}{5.0cm}' + '\n')
        fsum.write(r'\usepackage{fancyhdr}' + '\n')
        fsum.write(r'\pagestyle{fancy}' + '\n')
        fsum.write(r'\lhead{\leftmark}' + '\n')
        fsum.write(r'\rhead{\thepage}' + '\n')
        fsum.write(r'\cfoot{}' + '\n')
        fsum.write(r'\begin{document}' + '\n')
    else:
        fsum.write(r'\clearpage' + '\n')
        fsum.write(r'\newpage' + '\n')

    # Get some values from the fitter
    start = fitter.toas.first_MJD.value
    start_ymd = fitter.toas.first_MJD.to_value(format='iso')
    finish = fitter.toas.last_MJD.value
    finish_ymd = fitter.toas.last_MJD.to_value(format='iso')
    span = finish - start

    label = f"{psr} {'narrowband' if NB else 'wideband'}"
    # Write beginning header info
    fsum.write(r'\section*{PSR ' + label + '\markboth{' + label + '}{}}\n')
    
    try:
        who = check_output(['git','config','--get','user.name'], text=True).strip()
    except CalledProcessError:
        who = "anonymous user"

    when = time.strftime("%Y %b %d (%a) %H:%M:%S GMT", time.gmtime())
    fsum.write(f'Summary generated on {when} by {who}' + r'\\' + '\n')
    # print par file
    fsum.write(r'Input par file: ' + verb(parfile) + r'\\' + '\n')
    # print tim file directory
    rls_dir = fitter.toas.filename[0].rpartition('/')[0]
    fsum.write(r'Input tim file directory: ' + verb(rls_dir) + r'\\' + '\n')
    # print list of tim file names, limit two tim files per line
    fsum.write(r'Input tim files:' + "\n")
    fsum.write(r'\begin{itemize}' + "\n")
    if isinstance(fitter.toas.filename, str):
        tim_files = [fitter.toas.filename]
    else:
        tim_files = fitter.toas.filename
    for tf in tim_files:
        fsum.write(r'\item ' + verb(tf.split('/')[-1]) + '\n')
    fsum.write(r'\end{itemize}' + "\n")
    fsum.write('Span: %.1f years (%s -- %s)\\\\\n ' % (span/365.24,
        str(start_ymd).split(' ')[0], str(finish_ymd).split(' ')[0]))

    if NB:
        try:
            avg_dict = fitter.resids.ecorr_average(use_noise_model=True)
            mjdlist = np.sort(avg_dict['mjds'].value)
        except:
            log.warning("Cannot get epoch averaged residual MJDs, Epoch calculation will use all MJDs and may not be correct.")
            mjdlist = np.sort(fitter.toas.get_mjds().value)
    else:
        mjdlist = np.sort(fitter.toas.get_mjds().value)
    maxepoch = 6.5
    nepoch = 1
    m0 = mjdlist[0]
    for m in mjdlist:
        if m > m0+maxepoch:
            nepoch += 1
            m0 = m
    fsum.write('Epochs (defined as observations within %.1f-day spans): %d\\\\\n' % (maxepoch,nepoch))

    # Print what fitter was used:
    fsum.write('Wideband data: %s\n' %('No' if NB else 'Yes'))
    fsum.write('\\\\Fitter: %s\n' %(fitter.__class__.__name__))

    model = get_model(parfile)
    # Write out the timing model
    fsum.write(r'\subsection*{Timing model}' + '\n')
    fsum.write(r"This is the timing model as specified in the input par file, with no additional fitting." + '\\\\\n')
    fsum.write(r'\begin{verbatim}' + '\n')
    # Get the parfile lines
    parlines = model.as_parfile().split('\n')
    for l in parlines:
        if l.startswith('DMX'): continue
        fsum.write(l+"\n")
    fsum.write(r'\end{verbatim}' + '\n')

    # Write out the residual stats
    fsum.write(r'\subsection*{Residual stats}' + '\n')
    fsum.write(r'\begin{verbatim}' + '\n')
    rs_keys = rs_dict.keys()
    for k in rs_keys:
        l = "# WRMS(%s) = %.3f %s" %(k, rs_dict[k]['wrms'].value, rs_dict[k]['wrms'].unit)
        fsum.write(l + '\n')
        l = "#  RMS(%s) = %.3f %s" %(k, rs_dict[k]['rms'].value, rs_dict[k]['rms'].unit)
        fsum.write(l + '\n\n')
    fsum.write(r'\end{verbatim}' + '\n')

    # Write out DM residual stats if desired
    if not NB and dm_dict != None:
        fsum.write(r'\subsection*{DM Residual stats}' + '\n')
        fsum.write(r'\begin{verbatim}' + '\n')
        rs_keys = rs_dict.keys()
        for k in rs_keys:
            l = "# WRMS(%s) = %.6f %s" %(k, dm_dict[k]['wrms'].value, dm_dict[k]['wrms'].unit)
            fsum.write(l + '\n')
            l = "#  RMS(%s) = %.6f %s" %(k, dm_dict[k]['rms'].value, dm_dict[k]['rms'].unit)
            fsum.write(l + '\n\n')
        fsum.write(r'\end{verbatim}' + '\n')

    # Check pulsar name
    fsum.write(r'\subsection*{Pulsar name check in .par file}' + '\n')
    fsum.write('Name in .par file: %s\\\\\n' % (model.PSR.value))
    if model.PSR.value.startswith("B") or model.PSR.value.startswith("J"):
        fsum.write('OK: starts with B or J\\\\\n')
    else:
        fsum.write(alert('Warning: does not start with B or J')+'\\\\\n')
    if not os.path.basename(parfile).startswith(model.PSR.value):
        msg = f'Warning: parfile is called {verb(parfile)} but pulsar name is {model.PSR.value}'
        fsum.write(alert(msg) + r"\\" + "\n")
    fsum.write("\n")
    
    if hasattr(model, "PB") and model.PB.value is not None:
        porb = model.PB.value
    elif hasattr(model, "FB0") and model.FB0.value is not None:
        porb = (1/model.FB0.value)/86400.
    else:
        porb = 0
    # Write Epochs section
    fsum.write(r'\subsection*{Epochs near center of data span?}' + '\n')
    tmidspan = 0.5*(float(finish)+float(start))
    fsum.write('Middle of data span: midspan = %.2f\\\\\n' % (tmidspan))
    dtpepoch = float(model.PEPOCH.value)-tmidspan
    fsum.write('PEPOCH - midspan = $%.2f$ days = $%.2f$ years\\\\\n'  % ( dtpepoch, dtpepoch/365.24))
    if abs(dtpepoch) > 1:
        fsum.write(alert('PEPOCH is not at center of data span')+"\\\\\n")
    if param_check('TASC', fitter, check_enabled=True):
        dttasc = float(model.TASC.value)-tmidspan
        fsum.write('TASC - midspan = $%.2f$ days = $%.1f$ years\\\\\n'  % ( dttasc, dttasc/365.24))
        if abs(dttasc) > 1 + porb:
            fsum.write(alert('TASC is not at center of data span')+"\\\\\n")
    if param_check('T0', fitter, check_enabled=True):
        dtt0 = float(model.T0.value)-tmidspan
        fsum.write('TASC - midspan = $%.2f$ days = $%.1f$ years\\\\\n'  % ( dtt0, dtt0/365.24))
        if abs(dtt0) > 1 + porb:
            fsum.write(alert('T0 is not at center of data span')+"\\\\\n")
    fsum.write('\n')

    # Check for more than one jumped receiver
    fsum.write(r'\subsection*{Receivers and JUMPs}' + '\n')
    groups = set(np.array(resids.toas.get_flag_value('f')[0]))
    receivers = set([g.replace("_GUPPI","").replace("_GASP","").replace("_PUPPI","").replace("_ASP","").replace("_YUPPI","") for g in groups])
    jumped = []
    for p in model.params:
        if "JUMP" in p and "DM" not in p:
            jumped.append(getattr(model, p).key_value[0])
    if len(jumped)==0:
        log.warning("no JUMPs")
        jumped = ()
    nnotjumped = 0
    fsum.write('{\\setlength{\\topsep}{6pt}%\n\\setlength{\\partopsep}{0pt}%\n')  # starts a new environment
    fsum.write('\\begin{tabbing}\\hspace*{72pt}\\=\\kill\n')
    fsum.write('Receivers:\\\\[4pt]')
    for r in receivers:
        fsum.write('\n')
        fsum.write(r.replace("_","\\_"))
        if r in jumped:
            fsum.write('\\> JUMP')
        else:
            nnotjumped += 1
        fsum.write('\\\\')
    if len(receivers)>0:
        fsum.write('[4pt]')
    if nnotjumped==1:
        fsum.write('One non-JUMPed receiver.  Good.\\\\')
    else:
        fsum.write(alert('Warning: %d non-JUMPed receivers.' % (nnotjumped,))+ '\\\\')
    fsum.write('\end{tabbing}\n')
    fsum.write('}\n\n')   # ends environment started above.

    fsum.write(r'\subsection*{Frozen parameters all zero?}' + '\n')
    any_dodgy = False
    ignoring = []
    for p in model.params_ordered:
        pm = getattr(model, p)
        if (isinstance(pm, (pint.models.parameter.floatParameter, 
                            pint.models.parameter.maskParameter,
                            pint.models.parameter.MJDParameter,
                            pint.models.parameter.AngleParameter,
                           )) 
            and pm.frozen
            and pm.value is not None
            and pm.value != 0):
            if p in {"START", "FINISH", "POSEPOCH", "DMEPOCH", "PEPOCH", "TZRMJD", "DM", "DMX", "NTOA", "CHI2", "DMDATA", "TZRFRQ", "RNAMP", "RNIDX", "CHI2R", "TRES", "SWP"}:
                ignoring.append(p)
                continue
            skip = False
            for pfx in ["EFAC", "EQUAD", "TN", "ECORR", "DMEFAC", "DMEQUAD"]:
                if p.startswith(pfx):
                    ignoring.append(p)
                    skip = True
                    break
            if skip:
                continue
            any_dodgy = True
            fsum.write(f"Parameter {verb(p)} is " + alert(f"frozen at {pm.value}")+"\\\\\n")
    if ignoring:
        w = ', '.join([verb(i) for i in ignoring])
        fsum.write(f"Ignoring {w}\\\\\n")
    if not any_dodgy:
        fsum.write("Yes.\\\\\n")

    fsum.write(r'\subsection*{par file fully fit?}' + '\n')
    chi2_initial = fitter.resids_init.chi2
    chi2_final = fitter.resids.chi2
    chi2_decrease = chi2_initial-chi2_final
    fsum.write(f"par file initial $\\chi^2$: {chi2_initial}\\\\\n")
    fsum.write(f"par file final $\\chi^2$: {chi2_final}\\\\\n")
    fsum.write(f"Decrease: {chi2_decrease}\\\\\n")
    if abs(chi2_decrease) > 0.01:
        if chi2_decrease > 0:
            msg = f"par file $\\chi^2$ decreased by {chi2_decrease} during fitting, fitter has not fully converged"
        else:
            msg = f"par file $\\chi^2$ increased by {-chi2_decrease} during fitting, fitter has produced bogus result"
        log.warning(msg)
        fsum.write(f'\\\\ Warning: {alert(msg)}\\\\\n')
    else:
        fsum.write(f'\\\\ Fitting produces no major change, all is probably fine.\\\\\n')
    sigma_threshold = 0.1
    max_cs = 0
    changed = None
    for p in fitter.model.free_params:
        # FIXME: replicate compare_model here? run compare_model? maybe with low verbosity but capture log messages?
        pm = getattr(fitter.model, p)
        iv = getattr(fitter.model_init, p).value
        if iv is None:
            fsum.write(alert("WARNING:") + f" free parameter {verb(p)} is unset in input model.\\\\\n")
            continue
        fv = pm.value
        u = pm.uncertainty.value
        cs = (iv-fv)/u
        if abs(cs) >= abs(max_cs):
            max_cs = cs
            changed = p
        if abs(cs) > sigma_threshold:
            msg = f"parameter {verb(p)} changed from {iv} to {fv} ({cs:.2g} sigma) during fit."
            log.warn(msg)
            fsum.write(alert("WARNING: ") + msg + "\\\\\n")
    fsum.write(f"Largest parameter change during fit was {verb(changed)} by {max_cs:.2g} sigma.\\\\\n")
                   
    # Write out if reduced chi squared is close to 1
    fsum.write(r'\subsection*{Reduced $\chi^2$ close to 1.00?}' + '\n')
    chi2_0 = fitter.resids_init.chi2
    ndof_0 = fitter.resids_init.dof
    rchi= chi2_0/ndof_0
    fpp = scipy.stats.chi2(int(ndof_0)).sf(float(chi2_0))
    fsum.write('Reduced $\chi^2$ is %f/%d = %f (false positive probability %g)\n' % (chi2_0,ndof_0,rchi,fpp))
    if rchi<0.95 or rchi>1.05:
        # Eh. Not clear if this is useful given an FPP.
        fsum.write('\\\\ Warning: $\chi^2$ is far from 1.00\n')
    if 0.001<fpp<0.999:
        fsum.write('\\\\ False positive probability is believable\n')
    else:
        log.warning(f"Reduced chi-squared of {rchi} has unlikely false positive probability of {fpp}") 
        fsum.write('\\\\ Warning: '+alert('False positive probability not believable') + '\n')
    if fitter_noise is not None:
        fsum.write("\\\\\n")
        fsum.write("\\\\\n")
        fsum.write("After applying new noise model:\\\\\n")
        chi2_1 = fitter_noise.resids.chi2
        ndof_1 = fitter_noise.resids.dof
        rchi = chi2_1/ndof_1
        fpp = scipy.stats.chi2(int(ndof_1)).sf(float(chi2_1))
        fsum.write('New reduced $\chi^2$ is %f/%d = %f (false positive probability %g)\n' % (chi2_1,ndof_1,rchi,fpp))
        if 0.001<fpp<0.999:
            fsum.write('\\\\ New false positive probability is believable\n')
        else:
            log.warning(f"New reduced chi-squared of {rchi} has unlikely false positive probability of {fpp}") 
            fsum.write('\\\\ Warning: '+alert('New false positive probability not believable') + '\n')

    # Check EFACs, EQUADs, ECORRs:
    fsum.write(r'\subsection*{Error parameters reasonable?}' + '\n')
    any_bad_efac = []
    fsum.write(r"\begin{tabular}{l c}" + "\n")
    fsum.write(r"Parameter & value\\" + "\n")
    for p in sorted(model.params):
        pm = getattr(model, p)
        if p.startswith("EFAC") or p.startswith("DMEFAC"):
            val = f"{pm.value:.3f}"
            if not 0.8 < pm.value < 1.2:
                val = alert(val)
                any_bad_efac.append(p)
            fsum.write(verb(f"{p} {pm.key} {pm.key_value[0]}") + f" & {val}\\\\\n")
    fsum.write(r"\end{tabular}\\" + "\n")    
    if any_bad_efac:
        msg = f"Some EFACs seem very large or small, has something gone wrong?"
        efacs = f" {', '.join(verb(e) for e in any_bad_efac)}"
        fsum.write(alert(msg) + efacs + "\\\\\n")
        log.warning(msg + efacs)
    else:
        fsum.write("All EFACs seem reasonable.\\\\\n")
    fsum.write("\n")
    fsum.write(r"\begin{tabular}{l c c c}" + "\n")
    fsum.write(r"Parameter & value & TOA median & ratio\\" + "\n")
    prior = False
    large_ratio = False
    large_value = False
    for p in sorted(model.params):
        pm = getattr(model, p)
        if pm.value is None:
            value = np.nan
        else:
            value = pm.value
        if p.startswith("EQUAD") or p.startswith("ECORR"):
            unc = np.median(fitter.toas.table["error"][pm.select_toa_mask(fitter.toas)])
            ratio = value/unc
            r = f"{ratio:.2f}"
            ro = r
            if ratio>0.75: 
                large_ratio = True
                ro = alert(r)
            if pm.value>9.5:
                prior = True
                ro = alert(r)
            if pm.value>3:
                large_value = True
                ro = alert(r)
            fsum.write(verb(f"{p} {pm.key} {pm.key_value[0]}") + f" & {pm.value:.3f} $\\mu$s & "
                       f"{unc:.3f} $\\mu$s & {ro}" + "\\\\\n")
        if p.startswith("DMEQUAD"):
            unc = np.median(fitter.toas.get_dm_errors().to_value(pint.dmu)[pm.select_toa_mask(fitter.toas)])
            ratio = value/unc
            r = f"{ratio:.2f}"
            if ratio>0.75:
                large_ratio = True
                r = alert(r)

            fsum.write(verb(f"{p} {pm.key} {pm.key_value}") + f" & {value:.3g} dmu & "
                       f"{unc:.3g} dmu & {r}" + "\\\\\n")
    fsum.write(r"\end{tabular}\\" + "\n")
    if large_ratio:
        fsum.write(alert("EQUADs and DMEQUADs that are large compared to the uncertainties on the "
                   "relevant TOAs may be a sign of something strange.") + "\\\\\n")
    if large_value:
        fsum.write(alert("Any EQUAD or ECORR more than a few us is suspicious.")
                   + "\\\\\n")
    if prior:
        fsum.write(alert("The prior limit on EQUADs and ECORRs is 10 us; any value close to this is suspicious.")
                   + "\\\\\n")
    
    fsum.write(r'\subsection*{Error parameters agree with chains?}' + '\n')
    if fitter_noise is None:
        fsum.write("\n")
        fsum.write("Noise chains not available.\\\\\n")
    else:
        if hasattr(fitter.model, "created_time"):
            fsum.write(f"Par file created: {fitter.model.created_time}" + r"\\" + "\n")
        if hasattr(fitter.model, "file_mtime"):
            fsum.write(f"Par file modified/checked out of git: {fitter.model.file_mtime}" + r"\\" + "\n")
        if hasattr(fitter_noise.model, "noise_mtime"):
            fsum.write(f"Noise chains created: {fitter_noise.model.noise_mtime}" + r"\\" + "\n")
        any_bogus = False
        fsum.write(r"\begin{tabular}{l c c c c}" + "\n")
        fsum.write(r"Parameter & par value & chain value & ratio & TOA median\\" + "\n")
        for p in sorted(model.params):
            pm = getattr(model, p)
            for pfx in ["EQUAD", "ECORR", "EFAC", "DMEQUAD", "DMECORR", "TN"]: 
                if not p.startswith(pfx):
                    continue
                if pm.value is None:
                    break
                try:
                    pm_noise = getattr(fitter_noise.model, p)
                except AttributeError:
                    pm_noise_value = r"$\cdot$"
                    r = r"$\cdot$"
                else:
                    pm_noise_value = f"{pm_noise.value:.3g} {pm_noise.units}"
                    ratio = pm.value/pm_noise.value
                    r = f"{ratio:.2f}"
                    if not 0.9 < ratio < 1.1:
                        r = alert(r)
                        any_bogus = True
                if hasattr(pm, "select_toa_mask"):
                    name = f"{p} {pm.key} {pm.key_value[0] if pm.key_value else pm.key_value}"
                else:
                    name = p
                if p.startswith("DMEQUAD"):
                    unc = np.median(fitter.toas.get_dm_errors().to_value(pint.dmu)[pm.select_toa_mask(fitter.toas)])
                    median = f"{unc:.3g} {pm.units}"
                elif p.startswith("EQUAD") or p.startswith("ECORR"):
                    unc = np.median(fitter.toas.table["error"][pm.select_toa_mask(fitter.toas)])
                    median = f"{unc:.3g} {pm.units}"
                else:
                    median = ""
                fsum.write(f"{verb(name)} & {pm.value:.3g} {pm.units} & {pm_noise_value} & {r} & {median}" + "\\\\\n")
        fsum.write(r"\end{tabular}\\" + "\n")
        if any_bogus:
            fsum.write(alert("Some noise parameters (marked in bold) appear to be different "
                       "in the noise chains than in the par file.")+"\\\\\n")
        model_set_params = {p for p in model.params if getattr(model, p).value is not None}
        noise_model_set_params = {p for p in fitter_noise.model.params if getattr(fitter_noise.model, p).value is not None}
        par_not_noise = list(sorted(model_set_params - noise_model_set_params))
        if par_not_noise:
            fsum.write(alert("WARNING: the par file contains (a) parameter(s) missing from the post-noise model:"))
            fsum.write(", ".join(verb(p) for p in par_not_noise))
            fsum.write(r"\\" + "\n")
        noise_not_par = list(sorted(model_set_params - noise_model_set_params))
        if noise_not_par:
            fsum.write(alert("WARNING: the post-noise model contains (a) parameter(s) missing from the par file:"))
            fsum.write(", ".join(verb(p) for p in noise_not_par))
            fsum.write(r"\\" + "\n")
    
    # Get lines to write for F-tests
    if NB:
        hdrline = "%42s %7s %9s %5s %s" % ("", "RMS(us)", "Chi2", "NDOF", "Ftest")
    else:
        hdrline = "%42s %7s %9s %9s %5s %s" % ("", "RMS(us)", "DM RMS(pc cm^-3)", "Chi2", "NDOF", "Ftest")
    fsum.write(r'\subsection*{Parameter tests}' + '\n')
    if Ftest_dict is None:
        fsum.write("\n")
        fsum.write("F test results not available.\\\\\n")
    else:
        ftest_lines = get_Ftest_lines(Ftest_dict, fitter)
        # Write F-test results
        fsum.write("F-test results used PINT\n")
        fsum.write(r'\begin{verbatim}' + '\n')
        fsum.write(hdrline + '\n')
        for l in ftest_lines:
            fsum.write(l + '\n')
        fsum.write(r'\end{verbatim}' + '\n')
        # Write a summary of the F-test results and suggestions
        add_statement, remove_statement = summarize_Ftest(Ftest_dict, fitter, alpha = ALPHA)
        fsum.write(add_statement + "\n")
        fsum.write(remove_statement + "\n")

    # Write if there are bad DMX ranges

    # NOTE - CURRENTLY CANNOT DO THIS, NEED DMX CHECKER FIRST
    if False:
        fsum.write(r'\subsection*{Check for bad DMX ranges, less than 10\% bandwidth}' + '\n')
        """
        if not is_wideband:
            if len(baddmx)==0:
                fsum.write('No bad dmx ranges\\\\\n')
            else:
                fsum.write('Bad DMX ranges found, %d out of %d DMX ranges:\\\\\n' % (len(baddmx),ndmx))
                for l in baddmx:
                    fsum.write('{\\tt '+l+'}\\\\\n')
        else:
            fsum.write('No fractional bandwidth check for DMX ranges with wideband data!\\\\\n')
        """
        if not NB:
            fsum.write('No fractional bandwidth check for DMX ranges with wideband data!\\\\\n')

    # compare_models
    fsum.write(r'\subsection*{Comparison with previous model}' + '\n')
    if previous_parfile is None:
        fsum.write("\n")
        fsum.write("No previous par file specified.\\\\\n")
    else:
        fsum.write("\n")
        fsum.write(f'Current par file: {verb(parfile)}' + '\\\\\n')
        fsum.write(f'Previous par file: {verb(previous_parfile)}' + '\\\\\n')
        fsum.write("\n")
        model_copy = copy.deepcopy(model)
        previous_model = get_model(previous_parfile)
        model_copy.name = "Current"
        previous_model.name = "Previous"
        try:
            cm = model_copy.compare(previous_model,
                               verbosity='max',
                               nodmx=True,
                               threshold_sigma=3)
        except ValueError as e:
            fsum.write(f"WARNING: {verb('compare_models')} failed because of {verb(e)}\\\\\n")
        else:
            fsum.write(r"{\small" + "\n")
            fsum.write(r"\begin{verbatim}" + "\n")
            fsum.write("\n".join(interesting_compare_lines(cm)))
            fsum.write(r"\end{verbatim}" + "\n")
            fsum.write(r"}" + "\n")
            fsum.write("\n")
            
    # Check excision percentage, add text to summary
    if cuts_dict is not None:

        # info about cut flags
        ntoa = cuts_dict['cut']['good'][0]
        cutflag_vals = [cdv[1] for cdv in cuts_dict['cut'].values()]
        total_cuts = np.sum(cutflag_vals)-cuts_dict['cut']['good'][1] # do not count "good"
        cut_pct = 100.0 * float(total_cuts)/ntoa

        fsum.write(r'\subsubsection*{Excised TOAs by Cut Flag}' + '\n')
        fsum.write(f"Note: {cuts_dict['cut']['good'][1]} good TOAs remain out of {ntoa} total." + "\\\\\n")
        if cut_pct > 75.0:
            msg = f"Lots of TOAs have been excised ({round(cut_pct, 1)}\%)! Is that intended? See attached plots."
            fsum.write(alert(msg) + "\\\\\n")
        for cf in cuts_dict['cut'].keys():
            if cf != 'good':
                tot,cut = cuts_dict['cut'][cf]
                fsum.write('%s: %i \\\\\n' % (verb(cf), cut))
        if 'badfile' not in cuts_dict['cut'] and 'badtoa' not in cuts_dict['cut']:
            fsum.write('No TOAs have been manually excised, so no manual cut plot will be appended to the PDF.\\\\\n')

        # cuts per telescope
        fsum.write(r'\subsubsection*{Excised TOAs by Observatory}' + '\n')
        for tel in cuts_dict['tel'].keys():
            cutwarn = ""
            tot,cut = cuts_dict['tel'][tel]
            cut_pct = 100.0 * float(cut)/tot
            remain = tot-cut
            if cut_pct > 75.0: cutwarn = alert(f"{round(cut_pct,1)}\% cut!")
            fsum.write('%s: %i TOAs remain (%i cut; %i total). %s \\\\\n' % (verb(tel), remain, cut, tot, cutwarn))

        # cuts per frontend/backend combo
        problem_febes = []
        fsum.write(r'\subsubsection*{Excised TOAs by Frontend/Backend Combination}' + '\n')
        for febe in cuts_dict['f'].keys():
            cutwarn = ""
            tot,cut = cuts_dict['f'][febe]
            cut_pct = 100.0 * float(cut)/tot
            remain = tot-cut
            if cut_pct > 75.0: cutwarn = alert(f"{round(cut_pct,1)}\% cut!")
            fsum.write('%s: %i TOAs remain (%i cut; %i total). %s \\\\\n' % (verb(febe), remain, cut, tot, cutwarn))

            if float(remain)/ntoa < 0.05:
                problem_febes.append(verb(febe))

        if problem_febes:
            msg = alert("Very few TOAs; noise parameters may be poorly determined for: ")
            febe_str = ', '.join(problem_febes)
            fsum.write(msg + febe_str + "\\\\\n")

    # Write out software versions used
    fsum.write(r'\subsection*{Software versions used in timing\_analysis:}' + '\n')
    fsum.write('PINT: %s\\\\\n' % verb(pint.__version__))
    fsum.write('astropy: %s\\\\\n' % verb(astropy.__version__))
    fsum.write('numpy: %s\\\\\n' % verb(np.__version__))
    fsum.write('python: %s\\\\\n' % verb(sys.version))
    try:
        import enterprise
        fsum.write('enterprise: %s\\\\\n' % (enterprise.__version__))
    except ImportError as error:
        log.warning(str(error)+ ", cannot print enterprise version.")
    if False:
        # We didn't run PTMCMCSampler, no guarantee this is the relevant version
        try:
            import PTMCMCSampler
            fsum.write('PTMCMCSampler: %s\\\\\n' % (PTMCMCSampler.__version__))
        except ImportError as error:
            log.warning(str(error)+ ", cannot print PTMCMCSampler version.")
    if False:
        # We didn't actually use PSRCHIVE, so it's misleading to report a version here
        # If it's available from the TOAs we could report that
        try:
            # FIXME: this is the psrchive version on the notebook server, but we never use this version
            # Is there any point reporting it?
            psrchive_v = check_output(["psrchive", "--version"]).decode("utf-8")
            fsum.write('PSRCHIVE: %s\\\\\n' % (psrchive_v))
        except (ImportError, FileNotFoundError) as error:
            log.warning(str(error)+ ", cannot print PSRCHIVE version.")
    
    # Write out the plots - Assuming we have already made the summary plot previous to this
    # FIXME: why not make the summary plots here?
    # TODO Fix the plots...
    if NB:
        plot_file_list = sorted(glob.glob("%s*summary_plot_*_nb.*" % (model.PSR.value)))
    else:
        plot_file_list = sorted(glob.glob("%s*summary_plot_*_wb.*" % (model.PSR.value)))
    if not plot_file_list:
        raise IOError("Unable to find any summary plots to include in summary PDF!")
    for plot_file in plot_file_list:
        fsum.write(r'\begin{figure}[p]' + '\n')
        #fsum.write(r'\begin{center}' + '\n')
        #fsum.write(r'\vspace*{-2.0em}' + '\n')
        fsum.write(r'\centerline{\includegraphics[]{' + plot_file + '}}\n')
        #fsum.write(r'\end{center}' + '\n')
        fsum.write(r'\end{figure}' + '\n')
    nb_wb = "nb" if NB else "wb"
    noise_plot = f"{model.PSR.value}_noise_corner_{nb_wb}.pdf"
    noise_posterior_plots_list = sorted(glob.glob(f"{model.PSR.value}_noise_posterior_{nb_wb}_*.pdf"))

    if no_corner:
        if not noise_posterior_plots_list:
            # raise IOError("Unable to find any noise posterior plots to include in summary PDF!")
            log.info(f"Could not find existing N noise posterior plots, e.g. {model.PSR.value}_noise_posterior_{nb_wb}_N.pdf")
            fsum.write(f"Noise posterior plots not found.\\\\\n")
        else:
            for npost_plt in noise_posterior_plots_list:
                log.info(f"Including noise posterior plots {npost_plt}")
                fsum.write(r'\begin{figure}[p]' + '\n')
                fsum.write(r'\centerline{\includegraphics[width=\textwidth]{' + npost_plt + '}}\n')
                fsum.write(r'\end{figure}' + '\n')
    else:
        if os.path.exists(noise_plot):
            log.info(f"Including noise corner plot {noise_plot}")
            fsum.write(r'\begin{figure}[p]' + '\n')
            fsum.write(r'\centerline{\includegraphics[width=\textwidth]{' + noise_plot + '}}\n')
            fsum.write(r'\end{figure}' + '\n')
        else:
            log.info(f"Could not find noise corner plot {noise_plot}")
            fsum.write(f"Noise corner plot {verb(noise_plot)} not found.\\\\\n")

    # excision donut plot
    if cuts_dict is not None:
        if NB:
            excise_plot_list = sorted(glob.glob("*%s.nb_donut.png" % (model.PSR.value)))
        else:
            excise_plot_list = sorted(glob.glob("*%s.wb_donut.png" % (model.PSR.value)))
        if not excise_plot_list:
            raise IOError("Unable to find any donut plots to include in summary PDF!")
        for ex_plt in excise_plot_list:
            fsum.write(r'\begin{figure}[p]' + '\n')
            fsum.write(r'\centerline{\includegraphics[width=0.5\linewidth]{' + ex_plt + '}}\n')
            fsum.write(r'\end{figure}' + '\n')
        # freq spread vs. MJD plot
        if NB:
            freq_plot_list = sorted(glob.glob("*%s*excision_nb.png" % (model.PSR.value)))
        else:
            freq_plot_list = sorted(glob.glob("*%s*excision_wb.png" % (model.PSR.value)))
        if not freq_plot_list:
            raise IOError("Unable to find any freq vs. MJD plots to include in summary PDF!")
        for freq_plt in freq_plot_list:
            fsum.write(r'\begin{figure}[p]' + '\n')
            fsum.write(r'\centerline{\includegraphics[width=0.9\linewidth]{' + freq_plt + '}}\n')
            fsum.write(r'\end{figure}' + '\n')
        # manual excision plots
        if 'badtoa' in cuts_dict['cut'] or 'badfile' in cuts_dict['cut']:
            if NB:
                hl_plot_list = sorted(glob.glob("%s_manual_hl_nb.png" % (model.PSR.value)))
            else:
                hl_plot_list = sorted(glob.glob("%s_manual_hl_wb.png" % (model.PSR.value)))
            if not hl_plot_list:
                raise IOError("Unable to find any manual cut highlight plots to include in summary PDF!")
            for hl_plt in hl_plot_list:
                fsum.write(r'\begin{figure}[p]' + '\n')
                fsum.write(r'\centerline{\includegraphics[width=\linewidth]{' + hl_plt + '}}\n')
                fsum.write(r'\end{figure}' + '\n')

    if append is None:

        fsum.write(r'\end{document}' + '\n')
        fsum.close()

        try:
            check_call(['pdflatex','-interaction=batchmode', texfile])
        except CalledProcessError as e:
            log.warning(f"Latex run failed: {e}")

def write_if_changed(filename, contents):
    """Write contents to filename, touching the file only if it does not already contain them.
    
    Inputs:
    ----------
    filename [string]: Name of a text file.
    contents [string]: Sting to write to the file.
    """
    if os.path.exists(filename):
        if contents == open(filename).read():
            return
    with open(filename, "w") as f:
        f.write(contents)

def apply_cut_flag(toas, orig_inds, flagvalue, warn=False):
    """Apply appropriate cut flag to a selection of toas.

    Inputs:
    ----------
    toas [pint.TOA]: PINT TOA object
    orig_inds [list]: toas.orig_table indices to cut
    flagvalue [string]: String to apply to TOA line
    """
    for i in orig_inds:
        if 'cut' not in toas.orig_table[i]['flags']:
            toas.orig_table[i]['flags']['cut'] = flagvalue
        elif warn:
            flags = toas.orig_table[i]['flags']
            if 'chan' in flags:
                log.warning(f"Skipping TOA {flags['name']} (chan {flags['chan']}, subint {flags['subint']}) already cut: {flags['cut']}.")
            else:  # wb TOAs do not have a chan flag
                log.warning(f"Skipping TOA {flags['name']} (subint {flags['subint']}) already cut: {flags['cut']}.")

def apply_cut_select(toas,reason='???'):
    """Apply toa selection based on cut flags present.

    Inputs:
    ----------
    toas [pint.TOA]: PINT TOA object
    reason [string]: String 
    """
    n_origtoas = len(toas.orig_table)
    mask = np.array(['cut' not in x for x in toas.orig_table['flags']])
    if len(toas.table) == n_origtoas:
        n_remove = n_origtoas - np.sum(mask)
        log.info(f"Selecting {sum(mask)} TOAs out of {n_origtoas} ({n_remove} removed based on {reason}).")
    else:
        n_more = len(toas.table) - np.sum(mask)
        log.info(f"Selecting {sum(mask)} TOAs out of {n_origtoas} ({n_more} more removed based on {reason}).")
    toas.table = toas.orig_table[mask]
    toas.table = toas.table.group_by('obs')  # otherwise table.groups.keys gets clobbered; consider using separate toas object

def check_recentness_noise(tc):
    """Check whether the timing configuration points to the most recent noise run.

    Returns
    =======
    name of chains in use, sorted list of available chains
        Under normal circumstances the first of these should equal the last element
        of the second; that is, the name of the directory in use should match the
        name of the most recent available set of chains
    """
    if not tc.get_noise_dir():
        log.warning(f"Yaml file does not have a noise-dir field (or it is unset). Will check working directory.")
        return None, None

    d = os.path.abspath(tc.get_noise_dir())
    if glob.glob(os.path.join(d,"chain*.txt")):
        log.warning(f'Ignoring chains directly in {d}. Chains should be in a subdirectory of {os.path.split(d)[1]} called {tc.get_source()}_{tc.get_toa_type().lower()}')
    noise_runs = [os.path.dirname(os.path.dirname(os.path.abspath(p))) 
                  for p in sorted(glob.glob(os.path.join(d,
                                                    "..",
                                                    "????-??-*",
                                                    tc.get_source()+"_"+tc.get_toa_type().lower(),
                                                    "chain*.txt")))]
    used_chains = os.path.basename(d)
    available_chains = [os.path.basename(n) for n in noise_runs]
    
    if not noise_runs:
        log.warning(f'Looking for noise chains in given noise-dir ({d}), but does not follow current conventions (shared chains in /nanograv/share/20yr/noise-chains/<PSR B/J NAME>/<YYYY-MM-DD>/).')
        noise_runs = [os.path.dirname(os.path.abspath(p)) for p in sorted(glob.glob(os.path.join(d, tc.get_source()+"_"+tc.get_toa_type().lower()+"*", "chain*.txt")))]
        if len(noise_runs) > 0:
            if len(noise_runs) == 1:
                log.info(f'{len(noise_runs)} noise chain found in noise-dir ({d}).')
            else:
                log.info(f'{len(noise_runs)} noise chains found in noise-dir ({d}). Using first in sorted list.')
            used_chains = os.path.abspath(noise_runs[0])
            available_chains = [os.path.abspath(n) for n in noise_runs]
    
    if not noise_runs:
        log.warning('No chains found. Will search working directory and apply if found.')

    log.info(f"Using: {used_chains}")
    log.info(f"Available: {' '.join(available_chains)}")
    try:
        newest_chains = available_chains[-1]
    except IndexError:
        log.warning(f"No noise chains are available for this pulsar! Make sure to run the noise modeling.")
    else:
        if used_chains != newest_chains:
            log.warning(f"Using chains from {used_chains} but {newest_chains} is available")
    return used_chains, available_chains

def check_recentness_excision(tc):
    """Check whether the timing configuration points to the most recent set of excised TOAs.

    Returns
    =======
    name of excision run in use, sorted list of available excision runs
        Under normal circumstances the first of these should equal the last element
        of the second; that is, the name of the directory in use should match the
        name of the most recent available set of chains
    """
    if not tc.get_excised():
        log.warning(f"Yaml file does not have an excised-tim field (or it is unset).")
        return None, None

    e = os.path.abspath(tc.get_excised())
    d = os.path.dirname(e)
    excision_dirs = [os.path.dirname(os.path.dirname(os.path.abspath(p))) 
                  for p in sorted(glob.glob(os.path.join(d,
                                                         "..", "..",
                                                         "*.Outlier.*",
                                                         tc.get_source()+"."+tc.get_toa_type().lower(),
                                                         "*_excise.tim")))]
    used_excision = os.path.basename(os.path.dirname(d))
    available_excision = [os.path.basename(n) for n in excision_dirs]
    log.info(f"Using: {used_excision}")
    log.info(f"Available: {' '.join(available_excision)}")
    if used_excision != available_excision[-1]:
        log.warning(f"Using excision from {used_excision} but {available_excision[-1]} is available")
    return used_excision, available_excision


def no_ecorr_average(toas, resids, use_noise_model=True):
        """Create a dictionary to match those created by ecorr_avg for TOAs w/no-ecorr -- 
        to be used with EPTA data for consistent IPTA averaged plots. 
        
        Use a toa object and residaul object as input as input instead of just residual object (oh well)
        Also tell it pre/post fit.
        
        Arguments
        ---------
        toas[object]: PINT TOA object.
        resids[object]: PINT residuals object
        use_noise_model[boolean]: Default true.
        
        Returns a dictionary with the following entries:

          mjds           Average MJD for each segment

          freqs          Average topocentric frequency for each segment

          time_resids    Average residual for each asegment, time units

          noise_resids   Dictionary of per-noise-component average residual

          errors         Uncertainty on averaged residuals

          indices        List of lists giving the indices of TOAs in the original TOA table for each segment
        """
        
        # We already know that TOAs do not have ecorr. Identify the indices where that's true.
        inds= np.where(toas['pta']=='EPTA')[0]
        
        only_epta_toas = toas[inds]
        
        no_avg = {}
        no_avg["mjds"] = only_epta_toas.get_mjds()

        if use_noise_model:
            err = resids.model.scaled_toa_uncertainty(only_epta_toas)
        else:
            err = only_epta_toas.get_errors()

        # Weighted average of various quantities
        no_avg = {}
        no_avg["mjds"] = only_epta_toas.get_mjds()
        no_avg["freqs"] = only_epta_toas.get_freqs()
        no_avg["time_resids"] = resids.time_resids[inds]
        no_avg["noise_resids"] = {}
        for k in resids.noise_resids.keys():
            no_avg["noise_resids"][k] = resids.noise_resids[k][inds]

        # Uncertainties
        no_avg["errors"] = err
        
        # Indices back into original TOA list
        no_avg["indices"] = inds
        

        return no_avg
    



'''
Unit testing for par_checker.py

These tests are performed on modified versions of the
NANOGrav 12.5-year data set (version 3) parameter files
'''


import unittest
from astropy import log
import pint.models as model
from pint_pal.par_checker import *

log.setLevel("ERROR") # do not show PINT warnings here to avoid clutter




class ParCheckerTests(unittest.TestCase):
    """ par_checker.py testing class """

    def test_check_if_fit(self):
        parfile = "par/bad_values.par"
        psr_model = model.get_model(parfile)
        self.assertIsNone(check_if_fit(psr_model, "DMX_0002"))
        with self.assertRaises(ValueError):
            check_if_fit(psr_model, "DMX_0001")

    def test_check_name(self):
        """
        Check to see if the pulsar name is in the proper format,
        either J####[+/-]####
        or B####[+/-]##
        """
        parfile = "par/J1024-0719_NANOGrav_12yv4.gls.par"
        psr_model = model.get_model(parfile)
        self.assertIsNone(check_name(psr_model))

        parfile = "par/bad_values.par"
        psr_model = model.get_model(parfile)
        with self.assertWarns(Warning):
            check_name(psr_model)

    def test_check_spin(self):
        """
        Check to see if the spin and spindown parameters
        are fit for, and also cases for F2
        """

        # Check F2 for J1024-0719
        parfile = "par/J1024-0719_NANOGrav_12yv4.gls.par"
        psr_model = model.get_model(parfile)
        self.assertIsNone(check_spin(psr_model))

        # F2 has been inserted here, which should throw a warning
        parfile = "par/bad_values.par"
        psr_model = model.get_model(parfile)
        with self.assertWarns(Warning):
            check_spin(psr_model)

    def test_check_astrometry(self):
        """
        Check to see if astrometric parameters are in ecliptic coordinates
        and parallax is fit for
        """

        # Parallax has been removed
        parfile = "par/bad_values.par"
        psr_model = model.get_model(parfile)
        with self.assertRaises(ValueError):
            check_astrometry(psr_model)


    def test_check_binary(self):
        pass

    def test_check_jumps(self):
        pass



try:
    unittest.main(argv=[''], verbosity=2)
except SystemExit: #cleaner output below
    print



import numpy as np, os, json, itertools, time
from astropy import log
from astropy.time import Time

from enterprise.pulsar import Pulsar
from enterprise_extensions import models, model_utils
from enterprise_extensions import sampler as ee_sampler
import corner

import pint.models as pm
from pint.models.parameter import maskParameter
from pint.models.timing_model import Component

import matplotlib.pyplot as pl

import la_forge.core as co

from enterprise_extensions.sampler import group_from_params, get_parameter_groups
from enterprise_extensions import model_utils
from enterprise_extensions.empirical_distr import (EmpiricalDistribution1D,
                                                   EmpiricalDistribution2D)


def setup_sampling_groups(pta,
                          write_groups=True,
                          outdir='./'):
    """
    Sets sampling groups for PTMCMCSampler.
    The sampling groups can help ensure the sampler does not get stuck.
    The idea is to group parameters which are more highly correlated.

    Params
    ------
    pta: the enterprise pta object 
    write_groups: bool, write the groups to a file
    outdir: str, directory to write the groups to
    
    returns
    -------
    groups: list of lists of indices corresponding to parameter groups
    
    """
        
        # groups
    pnames = pta.param_names
    groups = get_parameter_groups(pta)
    # add per-backend white noise
    backends = np.unique([p[p.index('_')+1:p.index('efac')-1] for p in pnames if 'efac' in p])
    for be in backends:
        groups.append(group_from_params(pta,[be]))
    # group red noise parameters
    exclude = ['linear_timing_model','sw_r2','sw_4p39','measurement_noise',
            'ecorr_sherman-morrison', 'ecorr_fast-sherman-morrison']
    red_signals = [p[p.index('_')+1:] for p in list(pta.signals.keys())
                if not p[p.index('_')+1:] in exclude]
    rn_ct = 0
    for rs in red_signals:
        if len(group_from_params(pta,[rs])) > 0:
            rn_ct += 1
            groups.append(group_from_params(pta,[rs]))
    if rn_ct > 1:
        groups.append(group_from_params(pta,red_signals))
    # add cross chromatic groups
    if 'n_earth' in pnames or 'log10_sigma_ne' in pnames:
        # cross SW and chrom groups
        dmgp_sw = [idx for idx, nm in enumerate(pnames)
                if any([flag in nm for flag in ['dm_gp','n_earth', 'log10_sigma_ne']])]
        groups.append(dmgp_sw)
        if np.any(['chrom' in param for param in pnames]):
            chromgp_sw = [idx for idx, nm in enumerate(pnames)
                        if any([flag in nm for flag in ['chrom_gp','n_earth', 'log10_sigma_ne']])]
            dmgp_chromgp_sw = [idx for idx, nm in enumerate(pnames)
                            if any([flag in nm for flag in ['dm_gp','chrom','n_earth', 'log10_sigma_ne']])]
            groups.append(chromgp_sw)
            groups.append(dmgp_chromgp_sw)
    if np.any(['chrom' in param for param in pnames]):
        # cross dmgp and chromgp group
        dmgp_chromgp = [idx for idx, nm in enumerate(pnames)
                        if any([flag in nm for flag in ['dm_gp','chrom']])]
        groups.append(dmgp_chromgp)
    # everything
    groups.append([i for i in range(len(pnames))])
    # save list of params corresponding to groups
    if write_groups is True:
        with open(f'{outdir}/groups.txt', 'w') as fi:
            for group in groups:
                line = np.array(pnames)[np.array(group)]
                fi.write("[" + " ".join(line) + "]\n")
    # return the groups to be passed to the sampler
    return groups


def get_mean_large_likelihoods(core, N=10):
    '''
    Calculate the mean of the top N likelihood samples from the chain.
    This is an alternate to fixing the noise values in the timing model to
    the MAP or the median.

    Params
    ======
    core: la_forge.core object
    N: int, number of top likelihood samples to average
    Returns
    =======
    mean_data: np.array, mean of the top N likelihood samples
    '''
    chain = core.chain[core.burn:,:]
    lnlike_idx = core.params.index('lnlike')
    sorted_data = chain[chain[:, lnlike_idx].argsort()[::-1]]
    vals = np.mean(sorted_data[:N,:],axis=0)
    return {par: vals[p] for p, par in enumerate(core.params)}


def analyze_noise(
    chaindir="./noise_run_chains/",
    use_noise_point='mean_large_likelihood',
    likelihoods_to_average=50,
    burn_frac=0.25,
    save_corner=True,
    no_corner_plot=False,
    chaindir_compare=None,
    model_kwargs={},
    sampler_kwargs={},
):
    """
    Reads enterprise chain file; produces and saves corner plot; returns WN dictionary and RN (SD) BF

    Parameters
    ==========
    chaindir: path to enterprise noise run chain; Default: './noise_run_chains/'
    use_noise_point: point to use for noise analysis; Default: 'MAP'.
        Options: 'MAP', 'median', 'mean_large_likelihood',
        Note that the MAP is the the same as the maximum likelihood value when all the priors are uniform. 
    likelihoods_to_average: number of top likelihood samples to average; Default: 50
        Only applicable if use_noise_point is 'mean_large_likelihood'.
    burn_frac: fraction of chain to use for burn-in; Default: 0.25
    save_corner: Flag to toggle saving of corner plots; Default: True
    no_corner_plot: Flag to toggle saving of corner plots; Default: False
    chaindir_compare: path to noise run chain wish to plot in corner plot for comparison; Default: None

    Returns
    =======
    noise_core: la_forge.core object which contains noise chains and run metadata
    noise_dict: Dictionary of maximum a posterior noise values
    rn_bf: Savage-Dickey BF for achromatic RN for given pulsar
    """
    # get the default settings
    model_defaults, sampler_defaults = get_model_and_sampler_default_settings()
    # update with args passed in
    model_defaults.update(model_kwargs)
    sampler_defaults.update(sampler_kwargs)
    model_kwargs = model_defaults.copy()
    sampler_kwargs = sampler_defaults.copy()
    sampler = sampler_kwargs['sampler']
    likelihood = sampler_kwargs['likelihood']
    try:
        noise_core = co.Core(chaindir=chaindir)
    except:
        if os.path.isfile(chaindir):
            log.error(f"Could not load noise run from {chaindir}. Make sure the path is correct. " \
                      +"Also make sure you have an up-to-date la_forge installation. ")
            raise ValueError(f"Could not load noise run from {chaindir}. Check path and la_forge installation.")
        else:
            log.error(f"No noise runs found in {chaindir}. Make sure the path is correct.") 
            raise ValueError(f"Could not load noise run from {chaindir}. Check path.")
    if sampler == 'PTMCMCSampler':
        # standard burn ins
        noise_core.set_burn(burn_frac)
    else:
        noise_core.set_burn(burn_frac)
    chain = noise_core.chain[int(burn_frac * len(noise_core.chain)) :, :-4]
    psr_name = noise_core.params[0].split("_")[0]
    pars =  np.array([p for p in noise_core.params if p not in ['lnlike', 'lnpost', 'chain_accept', 'pt_chain_accept']])
    # if len(pars)+2 != chain.shape[1]:
    #     chain = chain[:, :len(pars)+2]
    
    # load in same for comparison noise model
    if chaindir_compare is not None:
        compare_core = co.Core(chaindir=chaindir)
        compare_core.set_burn(noise_core.burn)
        chain_compare = compare_core.chain[int(burn_frac * len(noise_core.chain)) :, :-4]
        pars_compare = np.array([p for p in compare_core.params if p not in ['lnlike', 'lnpost', 'chain_accept', 'pt_chain_accept']])
        # if len(pars_compare)+2 != chain_compare.shape[1]:
        #     chain_compare = chain_compare[:, :len(pars_compare)+2]

        psr_name_compare = pars_compare[0].split("_")[0]
        if psr_name_compare != psr_name:
            log.warning(
                f"Pulsar name from {chaindir_compare} does not match. Will not plot comparison"
            )
            chaindir_compare = None

            
    if save_corner and not no_corner_plot:
        pars_short = [p.split("_", 1)[1] for p in pars]
        log.info(f"Chain parameter names are {pars_short}")
        log.info(f"Chain parameter convention: {test_equad_convention(pars_short)}")
        if chaindir_compare is not None:
            # need to plot comparison corner plot first so it's underneath
            compare_pars_short = [p.split("_", 1)[1] for p in pars_compare]
            log.info(f"Comparison chain parameter names are {compare_pars_short}")
            log.info(
               f"Comparison chain parameter convention: {test_equad_convention(compare_pars_short)}"
            )
            # don't plot comparison if the parameter names don't match
            if compare_pars_short != pars_short:
                log.warning(
                   "Parameter names for comparison noise chains do not match, not plotting the compare-noise-dir chains"
                )
                chaindir_compare = None
            else:
                normalization_factor = (
                    np.ones(len(chain_compare))
                    * len(chain)
                    / len(chain_compare)
                )
                fig = corner.corner(
                    chain_compare,
                    color="orange",
                    alpha=0.5,
                    weights=normalization_factor,
                    labels=compare_pars_short,
                )
                # normal corner plot
                corner.corner(
                    chain, fig=fig, color="black", labels=pars_short
                )
        if chaindir_compare is None:
            corner.corner(chain, labels=pars_short)

        if "_wb" in chaindir:
            figname = f"./{psr_name}_noise_corner_wb.pdf"
        elif "_nb" in chaindir:
            figname = f"./{psr_name}_noise_corner_nb.pdf"
        else:
            figname = f"./{psr_name}_noise_corner.pdf"

        pl.savefig(figname)
        pl.savefig(figname.replace(".pdf", ".png"), dpi=300)

        pl.show()

    if no_corner_plot:

        from matplotlib.backends.backend_pdf import PdfPages

        if "_wb" in chaindir:
            figbase = f"./{psr_name}_noise_posterior_wb"
        elif "_nb" in chaindir:
            figbase = f"./{psr_name}_noise_posterior_nb"
        else:
            figbase = f"./{psr_name}_noise_posterior"

        pars_short = [p.split("_", 1)[1] for p in pars]
        log.info(f"Chain parameter names are {pars_short}")
        log.info(f"Chain parameter convention: {test_equad_convention(pars_short)}")
        if chaindir_compare is not None:
            # need to plot comparison corner plot first so it's underneath
            compare_pars_short = [p.split("_", 1)[1] for p in pars_compare]
            log.info(f"Comparison chain parameter names are {compare_pars_short}")
            log.info(
                f"Comparison chain parameter convention: {test_equad_convention(compare_pars_short)}"
            )
            # don't plot comparison if the parameter names don't match
            if compare_pars_short != pars_short:
                log.warning(
                    "Parameter names for comparison noise chains do not match, not plotting the compare-noise-dir chains"
                )
                chaindir_compare = None
            else:
                normalization_factor = (
                    np.ones(len(chain_compare))
                    * len(chain)
                    / len(chain_compare)
                )

        # Set the shape of the subplots
        shape = pars.shape[0]

        if "_wb" in chaindir:
            ncols = 4  # number of columns per page
        else:
            ncols = 3

        nrows = 5  # number of rows per page

        mp_idx = noise_core.map_idx
        param_medians = [noise_core.get_param_median(p) for p in noise_core.params if p not in ['lnlike', 'lnpost']]
        param_medians_dict = {p: noise_core.get_param_median(p) for p in noise_core.params if p not in ['lnlike', 'lnpost']}
        #mp_idx = np.argmax(chain[:, a])
        if chaindir_compare is not None:
            mp_compare_idx = compare_core.map_idx

        nbins = 20
        pp = 0
        for idx, par in enumerate(pars_short):
            j = idx % (nrows * ncols)
            if j == 0:
                pp += 1
                fig = pl.figure(figsize=(8, 11))

            ax = fig.add_subplot(nrows, ncols, j + 1)
            ax.hist(
                chain[:, idx],
                bins=nbins,
                histtype="step",
                color="black",
                label="Current",
            )
            ax.axvline(chain[:, idx][mp_idx], ls="--", color="black", label="MAP")
            if use_noise_point == 'mean_large_likelihood':
                lbl = "mean of 50 MLVs"
            if use_noise_point == 'MAP':
                lbl = "MAP"
            if use_noise_point == 'median':
                lbl = "median"
            ax.axvline(param_medians[idx], ls="--", color="green", label=lbl)
            if chaindir_compare is not None:
                ax.hist(
                    chain_compare[:, idx],
                    bins=nbins,
                    histtype="step",
                    color="orange",
                    label="Past",
                )
                ax.axvline(
                    chain_compare[:, idx][mp_compare_idx], ls="--", color="orange"
                )
            if "_wb" in chaindir:
                ax.set_xlabel(par, fontsize=8)
            else:
                ax.set_xlabel(par, fontsize=10)
            ax.set_yticks([])
            ax.set_yticklabels([])

            if j == (nrows * ncols) - 1 or idx == len(pars_short) - 1:
                pl.tight_layout()
                pl.savefig(f"{figbase}_{pp}.pdf")

        # Wasn't working before, but how do I implement a legend?
        # ax[nr][nc].legend(loc = 'best')
        pl.legend(loc="best")
        pl.show()
    
    if use_noise_point == 'MAP':
        noise_dict = noise_core.get_map_dict()
    elif use_noise_point == 'median':
        noise_dict = param_medians_dict
    elif use_noise_point == 'mean_large_likelihood':
        noise_dict = get_mean_large_likelihoods(noise_core, N=likelihoods_to_average)
    else:
        log.error(f"Invalid noise point {use_noise_point}. Must be 'MAP' or 'median' ")
        raise ValueError(f"Invalid noise point {use_noise_point}. Must be 'MAP' or 'median' or 'mean_large_likelihood' ")

    # Print bayes factor for red noise in pulsar
    rn_amp_nm = psr_name+"_red_noise_log10_A"
    rn_bf = model_utils.bayes_fac(noise_core(rn_amp_nm), ntol=1, logAmax=-11, logAmin=-20)[0]

    return noise_core, noise_dict, rn_bf


def model_noise(
    mo,
    to,
    using_wideband=False,
    resume=False,
    run_noise_analysis=True,
    wb_efac_sigma=0.25,
    base_op_dir="./",
    model_kwargs={},
    sampler_kwargs={},
    return_sampler_without_sampling=False,
):
    """
    Setup enterprise or discovery likelihood and perform Bayesian inference on noise model

    Parameters
    ==========
    mo: PINT (or tempo2) timing model
    to: PINT (or tempo2) TOAs
    using_wideband: Flag to toggle between narrowband and wideband datasets; Default: False
    resume: Flag to resume overwrite previous run or not.
    run_noise_analysis: Flag to toggle execution of noise modeling; Default: True
    noise_kwargs: dictionary of noise model parameters; Default: {}
    sampler_kwargs: dictionary of sampler parameters; Default: {}
    return_sampler: Flag to return the sampler object; Default: False
    
    Recommended to pass model_kwargs and sampler_kwargs from the config file.
    Default kwargs given by function `get_model_and_sampler_default_settings`.
    Import configuration parameters:
        likelihood: choose from ['enterprise', 'discovery']
            enterprise -- enterprise likelihood
            discovery -- various numpyro samplers with a discovery likelihood
        sampler: for enterprise choose from ['PTMCMCSampler','GibbsSampler']
             for discovery choose from  ['HMC', 'NUTS', 'HMC-GIBBS']

    Returns
    =======
    None or
    samp: sampler object
    """
    # get the default settings
    model_defaults, sampler_defaults = get_model_and_sampler_default_settings()
    # update with args passed in
    model_defaults.update(model_kwargs)
    sampler_defaults.update(sampler_kwargs)
    model_kwargs = model_defaults.copy()
    sampler_kwargs = sampler_defaults.copy()
    likelihood = sampler_kwargs['likelihood']
    sampler = sampler_kwargs['sampler']
    
    if not using_wideband:
        outdir = base_op_dir + mo.PSR.value + "_nb/"
    else:
        outdir = base_op_dir + mo.PSR.value + '_wb/'

    if os.path.exists(outdir) and run_noise_analysis and not resume:
        log.warning(
            f"A noise directory for pulsar {mo.PSR.value} already exists! "
            "Please rename the existing directory or specify a new location with "
            "base_op_dir. If you're trying to resume noise modeling, use "
            "resume=True with the existing directory. Skipping noise analysis."
        )
        return None
    elif os.path.exists(outdir) and (run_noise_analysis) and (resume):
        log.info(
            "A noise directory for pulsar {} already exists! Re-running noise modeling starting from previous chain".format(
                mo.PSR.value
            )
        )

    if not run_noise_analysis:
        log.info(
            "Skipping noise modeling. Change run_noise_analysis = True to run noise modeling."
        )
        return None


    # Create enterprise Pulsar object for supplied pulsar timing model (mo) and toas (to)
    log.info(f"Creating enterprise.Pulsar object from model with {mo.NTOA.value} toas...")
    e_psr = Pulsar(mo, to)
    ##########################################################
    ################     PTMCMCSampler      ##################
    ##########################################################
    if likelihood == "enterprise" and sampler == 'PTMCMCSampler':
        log.info(f"Setting up noise analysis with {likelihood} likelihood and {sampler} sampler for {e_psr.name}")
        # Setup a single pulsar PTA using enterprise_extensions
        # Ensure n_iter is an integer
        sampler_kwargs['n_iter'] = int(float(sampler_kwargs['n_iter']))

        if sampler_kwargs['n_iter'] < 1e4:
            log.warning(
            f"Such a small number of iterations with {sampler} is unlikely to yield accurate posteriors. STRONGLY recommend increasing the number of iterations to at least 5e4"
            )
        if not using_wideband:
            pta = models.model_singlepsr_noise(
                e_psr,
                white_vary=True,
                red_var=model_kwargs['inc_rn'], # defaults True
                is_wideband=False,
                use_dmdata=False,
                dmjump_var=False,
                wb_efac_sigma=wb_efac_sigma,
                tm_svd=True,
                # DM GP
                #dm_var=model_kwargs['inc_dmgp'],
                #dm_Nfreqs=model_kwargs['dmgp_nfreqs'],
                # CHROM GP
                #chrom_gp=model_kwargs['inc_chromgp'],
                #chrom_Nfreqs=model_kwargs['chromgp_nfreqs'],
                #chrom_gp_kernel='diag', # Fourier basis chromg_gp
                # DM SOLAR WIND
                #dm_sw_deter=model_kwargs['inc_sw_deter'],
                #ACE_prior=model_kwargs['ACE_prior'],
                # can pass extra signals in here
                #extra_sigs=model_kwargs['extra_sigs'],
            )
            pta.set_default_params({})
        else:
            pta = models.model_singlepsr_noise(
                e_psr,
                is_wideband=True,
                use_dmdata=True,
                white_vary=True,
                red_var=model_kwargs['inc_rn'],
                dmjump_var=False,
                wb_efac_sigma=wb_efac_sigma,
                ng_twg_setup=True,
            )
            dmjump_params = {}
            for param in mo.params:
                if param.startswith("DMJUMP"):
                    dmjump_param = getattr(mo, param)
                    dmjump_param_name = (
                        f"{pta.pulsars[0]}_{dmjump_param.key_value[0]}_dmjump"
                    )
                    dmjump_params[dmjump_param_name] = dmjump_param.value
            pta.set_default_params(dmjump_params)
        # set groups here
        groups = setup_sampling_groups(pta, write_groups=False, outdir=outdir)
        #######
        # setup sampler using enterprise_extensions
        if sampler_kwargs['emp_distribution'] is not None:
            try:
                log.info(f"Attempting to load chains for an empirical distributions from {sampler_kwargs['emp_distribution']}")
                core = co.Core(chaindir=sampler_kwargs['emp_distribution'])
            except:
                log.warning(f"Failed to load chains for empirical distributions from {sampler_kwargs['emp_distribution']}.\nCheck path. Need absolute path to chain directory with `pars.txt` and `chain_1.txt`. files")
                core = None
            try:
                if core is not None:
                    emp_dist = make_emp_distr(core)
                    log.info(f"Successfully created empirical distributions !!")
                    log.info("Setting up sampler ...")
            except:
                log.warning(f"Failed to create empirical distributions from successfully loaded directory.")
                emp_dist = None
        else:
            log.warning("Setting up sampler without empirical distributions... Consider adding one for faster sampling by adding `emp_distribution`: /<path to old noise chain>/<psr name>_nb to the `noise_run`->`inference` section of the config file.")
            emp_dist = None
        samp = ee_sampler.setup_sampler(pta,
                                        outdir=outdir,
                                        resume=resume,
                                        groups=groups,
                                        empirical_distr = emp_dist,
        )
        if emp_dist is not None:
            try:
                samp.addProposalToCycle(samp.jp.draw_from_empirical_distr, 50)
            except:
                log.warning("Failed to add draws from empirical distribution.")
        # Initial sample
        # try to initialize the sampler to the maximum likelihood value from a previous run
        # initialize to a random point if any points are missing
        x0 = get_init_sample_from_chain_path(pta, chaindir=sampler_kwargs['emp_distribution'])
        try:
            log_single_likelihood_evaluation_time(pta, sampler_kwargs)
        except:
            log.warning("Failed to time likelihood.")
        if not return_sampler_without_sampling:
            # Start sampling
            log.info("Beginnning to sample...")
            samp.sample(
                x0, sampler_kwargs['n_iter'], SCAMweight=30, AMweight=15, DEweight=50, #**sampler_kwargs
            )
            log.info("Finished sampling.")
    elif likelihood == "enterprise" and sampler == 'GibbsSampler':
        log.info(f"Setting up noise analysis with {likelihood} likelihood and {sampler} sampler for {e_psr.name}")
        raise NotImplementedError("GibbsSampler not yet implemented for enterprise likelihood")
    elif likelihood == "discovery":
        log.info(f"Setting up noise analysis with {likelihood} likelihood and {sampler} sampler for {e_psr.name}")
        raise NotImplementedError("Discovery likelihood not yet implemented")
    else:
        log.error(
            f"Invalid likelihood ({likelihood}) and sampler ({sampler}) combination." \
            + "\nCan only use enterprise with PTMCMCSampler or GibbsSampler."
        )
    if return_sampler_without_sampling:
        return samp


def convert_to_RNAMP(value):
    """
    Utility function to convert enterprise RN amplitude to tempo2/PINT parfile RN amplitude
    """
    return (86400.0 * 365.24 * 1e6) / (2.0 * np.pi * np.sqrt(3.0)) * 10**value


def add_noise_to_model(
    model,
    use_noise_point='mean_large_likelihood',
    burn_frac=0.25,
    save_corner=True,
    no_corner_plot=False,
    ignore_red_noise=False,
    using_wideband=False,
    rn_bf_thres=1e2,
    base_dir=None,
    compare_dir=None,
    return_noise_core=False,
):
    """
    Add WN, RN, DMGP, ChromGP, and SW parameters to timing model.

    Parameters
    ==========
    model: PINT (or tempo2) timing model
    use_noise_point: point to use for noise analysis; Default: 'mean_large_likelihood'.
        Options: 'MAP', 'median', 'mean_large_likelihood'
        Note that the MAP is the the same as the maximum likelihood value when all the priors are uniform.
        Mean large likelihood takes N of the largest likelihood values and then takes the mean of those. (Recommended).
    burn_frac: fraction of chain to use for burn-in; Default: 0.25
    save_corner: Flag to toggle saving of corner plots; Default: True
    ignore_red_noise: Flag to manually force RN exclusion from timing model. When False,
        code determines whether
    RN is necessary based on whether the RN BF > 1e3. Default: False
    using_wideband: Flag to toggle between narrowband and wideband datasets; Default: False
    base_dir: directory containing {psr}_nb and {psr}_wb chains directories; if None, will
        check for results in the current working directory './'.
    return_noise_core: Flag to return the la_forge.core object; Default: False

    Returns
    =======
    model: New timing model which includes WN and RN (and potentially dmgp, chrom_gp, and solar wind) parameters
    (optional)
    noise_core: la_forge.core object which contains noise chains and run metadata
    """

    # Assume results are in current working directory if not specified
    if not base_dir:
        base_dir = "./"

    chaindir_compare = compare_dir
    if not using_wideband:
        chaindir = os.path.join(base_dir, f"{model.PSR.value}_nb/")
        if compare_dir is not None:
            chaindir_compare = os.path.join(compare_dir, f"{model.PSR.value}_nb/")
    else:
        chaindir = os.path.join(base_dir, f"{model.PSR.value}_wb/")
        if compare_dir is not None:
            chaindir_compare = os.path.join(compare_dir, f"{model.PSR.value}_wb/")

    log.info(f"Using existing noise analysis results in {chaindir}")
    log.info("Adding new noise parameters to model.")
    if use_noise_point == 'mean_large_likelihood':
        log.info("Using mean of top 50 likelihood samples for noise parameters.")
    elif use_noise_point == 'MAP':
        log.info("Using maximum a posteriori values for noise parameters.")
    elif use_noise_point == 'median':
        log.info("Using median values for noise parameters.")
    noise_core, noise_dict, rn_bf = analyze_noise(
        chaindir=chaindir,
        use_noise_point=use_noise_point,
        likelihoods_to_average=50,
        burn_frac=burn_frac,
        save_corner=save_corner,
        no_corner_plot=no_corner_plot,
        chaindir_compare=chaindir_compare,
    )
    chainfile = chaindir + "chain_1.txt"
    try:
        mtime = Time(os.path.getmtime(chainfile), format="unix")
        log.info(f"Noise chains loaded from {chainfile} created at {mtime.isot}")
    except:
        chainfile = chaindir+"chain.nc"
        mtime = Time(os.path.getmtime(chainfile), format="unix")
        log.info(f"Noise chains loaded from {chainfile} created at {mtime.isot}")
        

    # Create the maskParameter for EFACS
    efac_params = []
    equad_params = []
    ecorr_params = []
    dmefac_params = []
    dmequad_params = []

    efac_idx = 1
    equad_idx = 1
    ecorr_idx = 1
    dmefac_idx = 1
    dmequad_idx = 1
    
    psr_name = list(noise_dict.keys())[0].split("_")[0]
    noise_pars = np.array(list(noise_dict.keys()))
    wn_dict = {key: val for key, val in noise_dict.items() if "efac" in key or "equad" in key or "ecorr" in key}
    for key, val in wn_dict.items():
        
        if "_efac" in key:

            param_name = key.split("_efac")[0].split(psr_name)[1][1:]

            tp = maskParameter(
                name="EFAC",
                index=efac_idx,
                key="-f",
                key_value=param_name,
                value=val,
                units="",
                convert_tcb2tdb=False,
            )
            efac_params.append(tp)
            efac_idx += 1

        # See https://github.com/nanograv/enterprise/releases/tag/v3.3.0
        # ..._t2equad uses PINT/Tempo2/Tempo convention, resulting in total variance EFAC^2 x (toaerr^2 + EQUAD^2)
        elif "_t2equad" in key:

            param_name = (
                key.split("_t2equad")[0].split(psr_name)[1].split("_log10")[0][1:]
            )

            tp = maskParameter(
                name="EQUAD",
                index=equad_idx,
                key="-f",
                key_value=param_name,
                value=10**val / 1e-6,
                units="us",
                convert_tcb2tdb=False,
            )
            equad_params.append(tp)
            equad_idx += 1

        # ..._tnequad uses temponest convention, resulting in total variance EFAC^2 toaerr^2 + EQUAD^2
        elif "_tnequad" in key:

            param_name = (
                key.split("_tnequad")[0].split(psr_name)[1].split("_log10")[0][1:]
            )

            tp = maskParameter(
                name="EQUAD",
                index=equad_idx,
                key="-f",
                key_value=param_name,
                value=10**val / 1e-6,
                units="us",
                convert_tcb2tdb=False,
            )
            equad_params.append(tp)
            equad_idx += 1

        # ..._equad uses temponest convention; generated with enterprise pre-v3.3.0
        elif "_equad" in key:

            param_name = (
                key.split("_equad")[0].split(psr_name)[1].split("_log10")[0][1:]
            )

            tp = maskParameter(
                name="EQUAD",
                index=equad_idx,
                key="-f",
                key_value=param_name,
                value=10**val / 1e-6,
                units="us",
                convert_tcb2tdb=False,
            )
            equad_params.append(tp)
            equad_idx += 1

        elif ("_ecorr" in key) and (not using_wideband):

            param_name = (
                key.split("_ecorr")[0].split(psr_name)[1].split("_log10")[0][1:]
            )

            tp = maskParameter(
                name="ECORR",
                index=ecorr_idx,
                key="-f",
                key_value=param_name,
                value=10**val / 1e-6,
                units="us",
                convert_tcb2tdb=False,
            )
            ecorr_params.append(tp)
            ecorr_idx += 1

        elif ("_dmefac" in key) and (using_wideband):

            param_name = key.split("_dmefac")[0].split(psr_name)[1][1:]

            tp = maskParameter(
                name="DMEFAC",
                index=dmefac_idx,
                key="-f",
                key_value=param_name,
                value=val,
                units="",
                convert_tcb2tdb=False,
            )
            dmefac_params.append(tp)
            dmefac_idx += 1

        elif ("_dmequad" in key) and (using_wideband):

            param_name = (
                key.split("_dmequad")[0].split(psr_name)[1].split("_log10")[0][1:]
            )

            tp = maskParameter(
                name="DMEQUAD",
                index=dmequad_idx,
                key="-f",
                key_value=param_name,
                value=10**val,
                units="pc/cm3",
                convert_tcb2tdb=False,
            )
            dmequad_params.append(tp)
            dmequad_idx += 1

    # Test EQUAD convention and decide whether to convert
    convert_equad_to_t2 = False
    if test_equad_convention(noise_dict.keys()) == "tnequad":
        log.info(
            "WN paramaters use temponest convention; EQUAD values will be converted once added to model"
        )
        convert_equad_to_t2 = True
        if np.any(["_equad" in p for p in noise_dict.keys()]):
            log.info("WN parameters generated using enterprise pre-v3.3.0")
    elif test_equad_convention(noise_dict.keys()) == "t2equad":
        log.info("WN parameters use T2 convention; no conversion necessary")

    # Create white noise components and add them to the model
    ef_eq_comp = pm.ScaleToaError()
    ef_eq_comp.remove_param(param="EFAC1")
    ef_eq_comp.remove_param(param="EQUAD1")
    ef_eq_comp.remove_param(param="TNEQ1")
    for efac_param in efac_params:
        ef_eq_comp.add_param(param=efac_param, setup=True)
    for equad_param in equad_params:
        ef_eq_comp.add_param(param=equad_param, setup=True)
    model.add_component(ef_eq_comp, validate=True, force=True)

    if len(dmefac_params) > 0 or len(dmequad_params) > 0:
        dm_comp = pm.noise_model.ScaleDmError()
        dm_comp.remove_param(param="DMEFAC1")
        dm_comp.remove_param(param="DMEQUAD1")
        for dmefac_param in dmefac_params:
            dm_comp.add_param(param=dmefac_param, setup=True)
        for dmequad_param in dmequad_params:
            dm_comp.add_param(param=dmequad_param, setup=True)
        model.add_component(dm_comp, validate=True, force=True)
    if len(ecorr_params) > 0:
        ec_comp = pm.EcorrNoise()
        ec_comp.remove_param("ECORR1")
        for ecorr_param in ecorr_params:
            ec_comp.add_param(param=ecorr_param, setup=True)
        model.add_component(ec_comp, validate=True, force=True)

    # Create red noise component and add it to the model
    log.info(f"The SD Bayes factor for red noise in this pulsar is: {rn_bf}")
    if (rn_bf >= rn_bf_thres or np.isnan(rn_bf)) and (not ignore_red_noise):

        log.info("Including red noise for this pulsar")
        # Add the ML RN parameters to their component
        rn_comp = pm.PLRedNoise()

        rn_keys = np.array([key for key, val in noise_dict.items() if "_red_" in key])
        rn_comp.RNAMP.quantity = convert_to_RNAMP(
            noise_dict[psr_name + "_red_noise_log10_A"]
        )
        rn_comp.RNIDX.quantity = -1 * noise_dict[psr_name + "_red_noise_gamma"]
        # Add red noise to the timing model
        model.add_component(rn_comp, validate=True, force=True)
    else:
        log.info("Not including red noise for this pulsar")
        
    # Check to see if dm noise is present
    dm_pars = [key for key in noise_pars if "_dm_gp" in key]
    if len(dm_pars) > 0:
        ###### POWERLAW DM NOISE ######
        if f'{psr_name}_dm_gp_log10_A' in dm_pars:
            #dm_bf = model_utils.bayes_fac(noise_core(rn_amp_nm), ntol=1, logAmax=-11, logAmin=-20)[0] 
            #log.info(f"The SD Bayes factor for dm noise in this pulsar is: {dm_bf}") 
            log.info('Adding Powerlaw DM GP noise as PLDMNoise to par file')
            # Add the ML RN parameters to their component
            dm_comp = pm.noise_model.PLDMNoise()
            dm_keys = np.array([key for key, val in noise_dict.items() if "_red_" in key])
            dm_comp.TNDMAMP.quantity = convert_to_RNAMP(
                noise_dict[psr_name + "_dm_gp_log10_A"]
            )
            dm_comp.TNDMGAM.quantity = -1 * noise_dict[psr_name + "_dm_gp_gamma"]
            ##### FIXMEEEEEEE : need to figure out some way to softcode this
            dm_comp.TNDMC.quantitity = 100
            # Add red noise to the timing model
            model.add_component(dm_comp, validate=True, force=True)
        ###### FREE SPECTRAL (WaveX) DM NOISE ######
        elif f'{psr_name}_dm_gp_log10_rho_0' in dm_pars:
            log.info('Adding Free Spectral DM GP as DMWaveXnoise to par file')
            raise NotImplementedError('DMWaveXNoise not yet implemented')

    # Check to see if higher order chromatic noise is present
    chrom_pars = [key for key in noise_pars if "_chrom_gp" in key]
    if len(chrom_pars) > 0:
        ###### POWERLAW CHROMATIC NOISE ######
        if f'{psr_name}_chrom_gp_log10_A' in chrom_pars:
            log.info('Adding Powerlaw CHROM GP noise as PLCMNoise to par file')
            # Add the ML RN parameters to their component
            chrom_comp = pm.noise_model.PLCMNoise()
            # chrom_keys = np.array([key for key, val in noise_dict.items() if "_chrom_gp_" in key])
            chrom_comp.TNCMAMP.quantity = convert_to_RNAMP(
                noise_dict[psr_name + "_chrom_gp_log10_A"]
            )
            chrom_comp.TNCMGAM.quantity = -1 * noise_dict[psr_name + "_chrom_gp_gamma"]
            ##### FIXMEEEEEEE : need to figure out some way to softcode this
            chrom_comp.TNCMC.quantitity = 100
            # Add red noise to the timing model
            model.add_component(chrom_comp, validate=True, force=True)
        ###### FREE SPECTRAL (WaveX) DM NOISE ######
        elif f'{psr_name}_chrom_gp_log10_rho_0' in chrom_pars:
            log.info('Adding Free Spectral CHROM GP as CMWaveXnoise to par file')
            raise NotImplementedError('CMWaveXNoise not yet implemented')
            
    # Check to see if solar wind is present
    sw_pars = [key for key in noise_pars if "sw_r2" in key]
    if len(sw_pars) > 0:
        log.info('Adding Solar Wind Dispersion to par file')
        all_components = Component.component_types
        noise_class = all_components["SolarWindDispersion"]
        noise = noise_class()  # Make the dispersion instance.
        model.add_component(noise, validate=False, force=False)
        # add parameters
        if f'{psr_name}_n_earth' in sw_pars:
            model['NE_SW'].quantity = noise_dict[f'{psr_name}_n_earth']
            model['NE_SW'].frozen = True
        if f'{psr_name}_sw_gp_log10_A' in sw_pars:
            sw_comp = pm.noise_model.PLSWNoise()
            sw_comp.TNSWAMP.quantity = convert_to_RNAMP(noise_dict[f'{psr_name}_sw_gp_log10_A'])
            sw_comp.TNSWAMP.frozen = True
            sw_comp.TNSWGAM.quantity = -1.*noise_dict[f'{psr_name}_sw_gp_gamma']
            sw_comp.TNSWGAM.frozen = True
            # FIXMEEEEEEE : need to figure out some way to softcode this
            sw_comp.TNSWC.quantity = 10
            sw_comp.TNSWC.frozen = True
            model.add_component(sw_comp, validate=False, force=True)
        if f'{psr_name}_sw_gp_log10_rho' in sw_pars:
            raise NotImplementedError('Solar Wind Dispersion free spec GP not yet implemented')


    # Setup and validate the timing model to ensure things are correct
    model.setup()
    model.validate()
    model.meta['noise_mtime'] = mtime.isot

    if convert_equad_to_t2:
        from pint_pal.lite_utils import convert_enterprise_equads

        model = convert_enterprise_equads(model)

    if not return_noise_core:
        return model
    if return_noise_core:
        return model, noise_core


def test_equad_convention(pars_list):
    """
    If (t2/tn)equad present, report convention used.
    See https://github.com/nanograv/enterprise/releases/tag/v3.3.0

    Parameters
    ==========
    pars_list: list of noise parameters from enterprise run (e.g. pars.txt)

    Returns
    =======
    convention_test: t2equad/tnequad/None
    """
    # Test equad convention
    t2_test = np.any(["_t2equad" in p for p in pars_list])
    tn_test = np.any([("_tnequad" in p) or ("_equad" in p) for p in pars_list])
    if t2_test and not tn_test:
        return "t2equad"
    elif tn_test and not t2_test:
        return "tnequad"
    else:
        log.warning(
            "EQUADs not present in parameter list (or something strange is going on)."
        )
        return None


def get_init_sample_from_chain_path(pta, chaindir=None, json_path=None):
    """
    Get the initial sample from a chain directory or json file path.
    If parameters are missing, draw randomly from the prior
    Parameters
    ==========
    pta: enterprise PTA object
    chaindir: path to chain directory
    json_path: path to json file containing starting point
    Returns
    =======
    x0: initial sample
    """
    try:
        if chaindir is not None:
            log.info(f"Attempting to initialize sampler from MAP of chain directory {chaindir}")
            core = co.Core(chaindir)
            starting_point = core.get_map_dict()
            x0_dict = {}
            for prior, par_name in zip(pta.params, pta.param_names):
                if par_name in starting_point.keys():
                    x0_dict.update({par_name: starting_point[par_name]})
                else:
                    x0_dict.update({par_name:  prior.sample()})
            x0 = np.hstack([x0_dict[p] for p in pta.param_names])
        elif json_path is not None:
            with open(json_path, 'r') as fin:
                starting_point = json.load(fin)
            x0_dict = {}
            for prior, par_name in zip(pta.params, pta.param_names):
                if par_name in starting_point.keys():
                    x0_dict.update({par_name: starting_point[par_name]})
                else:
                    x0_dict.update({par_name:  prior.sample()})
            x0 = np.hstack([x0_dict[p] for p in pta.param_names])
        else:
            x0 = np.hstack([p.sample() for p in pta.params])
    except:
        x0 = np.hstack([p.sample() for p in pta.params])
        x0_dict = None
        log.warning(
            f"Unable to initialize sampler from chain directory or json file. Drawing random initial sample."
            )
    return x0

def make1d(par, samples, bins=None, nbins=81):
    if bins is None:
        bins = np.linspace(min(samples), max(samples), nbins)
        
    return EmpiricalDistribution1D(par, samples, bins)

def make2d(pars, samples, bins=None, nbins=81):
    idx = [0,1]
    if bins is None:
        bins = [np.linspace(min(samples[:, i]), max(samples[:, i]), nbins) for i in idx]
    return EmpiricalDistribution2D(pars, samples.T, bins)

def make_emp_distr(core):
    """
    Make empirical distributions for all parameters in core.
    Parameters
    ==========
    core: enterprise_extensions.core.Core object

    Returns
    =======
    dists: list of EmpiricalDistribution1D and EmpiricalDistribution2D objects
    """
    types = ['dm_gp', 'chrom_gp', 'red_noise', 'ecorr', 'chrom_s1yr', 'dm_s1yr', 'exp',]
    # made 1d hist for everything
    dists = [make1d(par, core(par)) for par in core.params[:-4] if 'chrom_gp_idx' not in par]
    # get list of parameters minus chrom_gp_idx cuz this prior is weird.
    params = [p for p in core.params if 'chrom_gp_idx' not in p]
    groups = {ii: [par for par in params if ii in par] for ii in types}
    # make 2ds for various related parameter subgroups
    for group in groups.values():
        _ = [dists.append(make2d(pars,core(list(pars)))) for pars in list(itertools.combinations(group,2)) if len(group)>1]
    # make 2d cross groups
    _ = [[dists.append(make2d([ecr, dm], core([ecr, dm]))) for ecr in groups['ecorr']] for dm in groups['dm_gp']]
    _ = [[dists.append(make2d([dm, chrom], core([dm, chrom]))) for dm in groups['dm_gp']] for chrom in groups['chrom_gp']]
    
    return dists

def log_single_likelihood_evaluation_time(pta, sampler_kwargs):
    """
    Log the time it takes to evaluate the likelihood once.
    """
    log.info("Building the enterprise likelihood and estimating evaluation time...")
    x1 = [[p.sample() for p in pta.params] for _ in range(11)]
    pta.get_lnlikelihood(x1[0])
    start_time = time.time()
    [pta.get_lnlikelihood(x1[i]) for i in range(1,11)]
    end_time = time.time()
    slet = (end_time-start_time)/10
    log.info(f"Single likelihood evaluation time is approximately {slet:.1e} seconds")
    log.info(f"4 times {sampler_kwargs['n_iter']} likelihood evaluations will take approximately: {4*slet*float(sampler_kwargs['n_iter'])/3600/24:.2f} days")



def get_model_and_sampler_default_settings():
    model_defaults = {
        # white noise
        'inc_wn': True, 
        'tnequad': True,
        # acrhomatic red noise
        'inc_rn': True,
        'rn_psd': 'powerlaw',
        'rn_nfreqs': 30,
        # dm gp
        'inc_dmgp': False,
        'dmgp_psd': 'powerlaw',
        'dmgp_nfreqs': 100,
        # higher order chromatic gp
        'inc_chromgp': False,
        'chromgp_psd': 'powerlaw',
        'chromgp_nfreqs': 100,
        'chrom_idx': 4,
        'chrom_quad': False,
        # solar wind
        'inc_sw_deter': False,
        # GP perturbations ontop of the deterministic model
        'inc_swgp': False,
        'ACE_prior': False,
        # 
        'extra_sigs': None,
        # misc
        'tm_svd': True
        }
    sampler_defaults = {
        'likelihood': 'enterprise',
        'sampler': 'PTMCMCSampler',
        # ptmcmc kwargs
        'n_iter': 2.5e5,
        'emp_distribution': None,
        # numpyro kwargs
        'num_steps': 25,
        'num_warmup': 500,
        'num_samples': 2500,
        'num_chains': 4,
        'chain_method': 'parallel',
        'max_tree_depth': 5,
        'dense_mass': False,
        }
    return model_defaults, sampler_defaults



"""
This module provides tools for round-tripping yamls and adding fields/blocks
if desired; ruyamel.yaml preserves order of existing fields and comments.
"""

from ruamel.yaml import YAML
import argparse
import glob
from astropy import log
import numpy as np
from pint_pal import config
import os

yaml = YAML()
RELEASE = f'/nanograv/timing/releases/15y/toagen/releases/{config.LATEST_TOA_RELEASE}/' 

def fix_toa_info(yaml_file,current_release=RELEASE,overwrite=True,extension='fix'):
    """Checks/fixes tim-directory, toas, toa-type from existing yaml; writes new one.

    Parameters
    ==========
    yaml_file: str, input file
    current_release: str, optional
        desired tim-directory (current release by default)
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    """
    file_only = yaml_file.split('/')[-1]
    source,toa_type,init_ext = file_only.split('.')

    # ASP/GASP files don't currently have nb/wb in them, so all are nb. GAH!
    # This is why I'm removing {toa-type} from the glob for now
    # Determine TOAs available for the source in the latest release
    release_toas = [gg.split('/')[-1] for gg in glob.glob(f'{current_release}{source}*.tim')]
    if toa_type == 'nb':
        release_toas = [x for x in release_toas if 'wb' not in x]
    elif toa_type == 'wb':
        release_toas = [x for x in release_toas if 'wb' in x]
    else:
        raise ValueError(f'Wrong file format (does not include nb/wb): {yaml_file}')

    release_toas.sort()
    changes = []
    config = read_yaml(yaml_file)

    # Check toa-type; currently case-sensitive
    if toa_type.upper() == config.get('toa-type'):
        pass
    else:
        log.warning(f'{source} yaml filename is not consistent with toa-type; check this!')

    # Check tim-directory and fix if necessary
    yamltimdir = config.get('tim-directory')
    log.debug(f'{source} tim-directory')
    log.debug(f'yaml: {yamltimdir}')
    log.debug(f'want: {current_release}')
    if not yamltimdir:
        log.error(f'{source} has no tim-directory listed. WTF.')
    elif yamltimdir == current_release:
        log.info(f'{source} tim-directory matches latest release.')
    else:
        log.warning(f'{source} tim-directory does not match latest release; fixing it.')
        config['tim-directory'] = current_release
        changes.append('tim-directory')

    # Check toas and fix if necessary
    yamltoas = config.get('toas')
    yamltoas.sort()
    log.debug(f'{source} toas')
    log.debug(f'yaml: {yamltoas}')
    log.debug(f'want: {release_toas}')
    if not yamltoas: 
        log.error(f'{source} has no {toa_type} toas listed. WTF.')
    elif yamltoas == release_toas:
        log.info(f'{source} current {toa_type} toas are all in use.')
    else:
        log.warning(f'{source} {toa_type} toas do not match those available; updating.')
        config['toas'] = release_toas
        changes.append('toas')

    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    # If there were changes, write an updated yaml file
    if not changes:
        log.info(f'{yaml_file} TOA info up to date; no new file will be written.')
        pass
    else:
        log.info(f'{source} changes were made to: {changes}')
        write_yaml(config, out_yaml)

def add_niterations(yaml_file,overwrite=True,extension='fix',insert_after='fitter'):
    """Adds n-iterations field to yaml file

    Parameters
    ==========
    yaml_file: str, input file
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    insert_after: str, optional
        field after which to insert this field in the yaml
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    if not config.get('n-iterations'):
        # n-iterations field goes after fitter by default (insert_after)
        insert_ind = list(config).index(insert_after)+1
        config.insert(insert_ind,'n-iterations',1)

        log.info(f'Adding n-iterations to {out_yaml}.')
        write_yaml(config, out_yaml)
    else:
        log.info(f'{yaml_file} already contains n-iterations field.')

def add_block_field(yaml_file,block_key,key,value,overwrite=True,extension='fix'):
    """Add key/value to specified block in the yaml

    Parameters
    ==========
    yaml_file: str, input file
    block_key: str, yaml block to which key/value should be added
    key: str, yaml field name
    value: variable type, set value associated with key
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    #d.keys().index(k) if we eventually want to add one key after another
    if key in config[block_key]:
        config[block_key][key] = value
        log.info(f'Config {block_key}/{key} already exists in {yaml_file}; setting it to {value}.')
    else:
        config[block_key][key] = value
        log.info(f'Adding {block_key}/{key} to {yaml_file}; setting it to {value}.')

    write_yaml(config, out_yaml)

def add_noise_block(yaml_file,overwrite=True,extension='fix',insert_after='bipm',
        noise_dir=None):
    """Adds noise block to yaml file

    Parameters
    ==========
    yaml_file: str, input file
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    insert_after: str, optional
        field after which to insert this block in the yaml
    noise_dir: str, optional
        base directory where existing noise results can be found
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    if not config.get('noise'):
        # noise block goes after bipm by default (insert_after)
        insert_ind = list(config).index(insert_after) + 1
        noise_block = {'results-dir':noise_dir}
        config.insert(insert_ind,'noise',noise_block,'control noise runs, apply results')
        log.info(f'Adding standard noise block to {out_yaml}.')
        write_yaml(config, out_yaml)
    else:
        log.info(f'{yaml_file} already contains noise block.')

def add_dmx_block(yaml_file,overwrite=True,extension='fix',insert_after='noise'):
    """Adds dmx block to yaml file

    Parameters
    ==========
    yaml_file: str, input file
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    insert_after: str, optional
        field after which to insert this block in the yaml
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    if not config.get('dmx'):
        # dmx block goes after noise by default (insert_after)
        insert_ind = list(config).index(insert_after) + 1
        dmx_block = {'ignore-dmx':False,'fratio':1.1,'max-sw-delay':0.1,'custom-dmx':[]}
        config.insert(insert_ind,'dmx',dmx_block,'control dmx windowing/fixing')
        log.info(f'Adding standard dmx block to {out_yaml}.')
        write_yaml(config, out_yaml)
    else:
        log.info(f'{yaml_file} already contains dmx block.')

def add_outlier_block(yaml_file,overwrite=True,extension='fix',insert_after='dmx'):
    """Adds outlier block to yaml file

    Parameters
    ==========
    yaml_file: str, input file
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    insert_after: str, optional
        field after which to insert this block in the yaml
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    if not config.get('outlier'):
        # outlier block goes after dmx by default (insert_after)
        insert_ind = list(config).index(insert_after) + 1
        outlier_block = {'method':'gibbs','n-burn':1000,'n-samples':20000}
        config.insert(insert_ind,'outlier',outlier_block,'control outlier analysis runs')
        log.info(f'Adding standard outlier block to {out_yaml}.')
        write_yaml(config, out_yaml)
    else:
        log.info(f'{yaml_file} already contains outlier block.')

def add_results_block(yaml_file,overwrite=True,extension='fix',insert_before='ignore',
        set_to_current=True):
    """Adds intermediate-results block to yaml file

    Parameters
    ==========
    yaml_file: str, input file
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    insert_before: str, optional
        field before which to insert this block in the yaml
    set_to_current: bool, optional
        set noise-dir and excised-tim approrpiately for current results available
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    # current settings
    if set_to_current:
        noise_dir = f"/nanograv/share/15yr/timing/intermediate/noise-chains/"
        excised_timfile = f"{config.get('source')}.{config.get('toa-type').lower()}_excise.tim"
        excised_timpath = f"/nanograv/share/15yr/timing/intermediate/excised-tims/"
        excised_tim = os.path.join(excised_timpath,excised_timfile)
        if not os.path.exists(excised_tim):
            log.warning(f"Excised tim file does not yet exist: {excised_timfile}")
            #excised_tim = None # meh, add it anyway
    else:
        noise_dir = None
        excised_tim = None

    if not config.get('intermediate-results'):
        # results block goes before ignore by default (insert_before)
        insert_ind = list(config).index(insert_before) 
        results_block = {'noise-dir':noise_dir,'excised-tim':excised_tim}
        config.insert(insert_ind,'intermediate-results',results_block,'use results from previous runs')
        log.info(f'Adding intermediate results block to {out_yaml}.')
        write_yaml(config, out_yaml)
    else:
        log.info(f'{yaml_file} already contains noise block.')

def add_check_block(yaml_file,overwrite=True,extension='fix',insert_after='ignore'):
    """Adds check block to yaml file

    Parameters
    ==========
    yaml_file: str, input file
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    insert_after: str, optional
        field after which to insert this block in the yaml
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    if not config.get('check'):
        # check block goes after ignore by default (insert_after)
        insert_ind = list(config).index(insert_after) + 1
        check_block = {
            'toa-outliers':None,
            'dmx-outliers':None,
            'unusual-params':None,
            'other':None,
            'cleared':False,
            }
        config.insert(insert_ind,'check',check_block,'check before final')
        log.info(f'Adding standard check block to {out_yaml}.')
        write_yaml(config, out_yaml)
    else:
        log.info(f'{yaml_file} already contains check block.')

def check_cleared(yaml_file,overwrite=True,extension='fix'):
    """Assigns 'cleared' status based on other check block fields

    Parameters
    ==========
    yaml_file: str, input file
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    if not config.get('check'):
        log.warning(f'{yaml_file} check block does not exist.')
    else:
        cleared = True
        for v in config.get('check').keys():  
            if not config['check'][v]:
                pass
            else:
                log.info(f'{yaml_file} not clear due to {v}.')
                cleared = False
                break

        if cleared:
            log.info(f'{yaml_file} check is cleared.')
            config['check']['cleared'] = True
            write_yaml(config, out_yaml)

def curate_comments(yaml_file,overwrite=True,extension='fix'):
    """Standardizes info comments on specific yaml fields

    Parameters
    ==========
    yaml_file: str, yaml filename
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    # currently assumes these fields exist, should check explicitly
    config.yaml_add_eol_comment("parameters not included here will be frozen",'free-params')
    config.yaml_add_eol_comment("toa excision",'ignore')

    toplevel_keys = dict(config).keys()
    ignore_keys = dict(config['ignore']).keys()
    dmx_keys = dict(config['dmx']).keys()

    if config.get('ignore').get('bad-toa'): 
        try:
            config['ignore'].yaml_add_eol_comment("designated by [name,chan,subint]",'bad-toa')
        except:
            pass
    else:
        if 'bad-toa' in ignore_keys:
            log.info('bad-toa field exists, is not set.')
        else:
            log.warning('bad-toa field does not exist.')

    if config.get('ignore').get('bad-range'):
        try:
            config['ignore'].yaml_add_eol_comment("designated by [mjd_start,mjd_end,(backend optional)]",'bad-range')
        except:
            pass
    else:
        if 'bad-range' in ignore_keys:
            log.info('bad-range field exists, is not set.')
        else:
            log.warning('bad-range field does not exist.')

    if config.get('ignore').get('bad-file'):
        try:
            config['ignore'].yaml_add_eol_comment("designated by basename string {backend}_{mjd}_{source}",'bad-file')
        except:
            pass
    else:
        if 'bad-file' in ignore_keys:
            log.info('bad-file field exists, is not set.')
        else:
            log.warning('bad-file field does not exist.')

    config['dmx'].yaml_add_eol_comment("finer binning when solar wind delay > threshold (us)",'max-sw-delay')
    config['dmx'].yaml_add_eol_comment("designated by [mjd_low,mjd_hi,binsize]",'custom-dmx')
    write_yaml(config, out_yaml)

def set_field(yaml_file,field,value,overwrite=True,extension='fix'):
    """Sets yaml field to provided value

    Parameters
    ==========
    yaml_file: str, yaml filename
    field: str, valid yaml key 
    value: str/list/dict, desired value
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    """
    config = read_yaml(yaml_file)
    out_yaml = get_outfile(yaml_file,overwrite=overwrite,extension=extension)

    valid_keys = ['source','par-directory','tim-directory','timing-model',
            'compare-model','toas','free-params','free-dmx','toa-type','fitter',
            'n-iterations','ephem','bipm','noise','results-dir','dmx','ignore-dmx',
            'fratio','max-sw-delay','custom-dmx','method','n-burn','n-samples',
            'intermediate-results','noise-dir',
            'compare-noise-dir','excised-tim','no-corner','ignore','mjd-start',
            'mjd-end','snr-cut','bad-toa','bad-range','bad-file','changelog',
            'toa-outliers','dmx-outliers','unusual-params','other','cleared']

    if field not in valid_keys:
        log.warning(f'Provided field ({field}) not valid.')
    else:
        if field == 'ephem' and isinstance(value,str):
            config['ephem'] = value
        elif field == 'mjd-end' and isinstance(value,float):
            config['ignore']['mjd-end'] = value
        elif field == 'timing-model' and isinstance(value,str):
            config['timing-model'] = value
        elif field == 'compare-model' and isinstance(value,str):
            config['compare-model'] = value
        # changelog appends value to existing
        elif field == 'changelog' and isinstance(value,str):
            config['changelog'].append(value)
        elif field == 'noise-dir' and isinstance(value,str):
            config['intermediate-results']['noise-dir'] = value
        elif field == 'compare-noise-dir' and isinstance(value,str):
            config['intermediate-results']['compare-noise-dir'] = value
        elif field == 'no-corner' and isinstance(value,bool):
            config['intermediate-results']['no-corner'] = value
        elif field == 'excised-tim' and isinstance(value,str):
            config['intermediate-results']['excised-tim'] = value
        elif field == 'method' and isinstance(value,str):
            config['outlier']['method'] = value
        else:
            log.error(f'Provided field ({field}) is valid, but not yet implemented in set_field(); doing nothing.')

    write_yaml(config, out_yaml)

def fix_badfile(yaml_file):
    """Suggests bad-file replacements (if strings) with one-element lists

    Parameters
    ==========
    yaml_file: str, yaml filename
    """
    config = read_yaml(yaml_file)
    try:
        n_be = len(config['ignore']['bad-file'])
        log.info(f'{yaml_file}: {n_be} bad-file entries to fix...')
        for i in range(n_be):
            be_field = config['ignore']['bad-file'][i]
            if isinstance(be_field,str):
                print(f'  - [{be_field}]')
    except TypeError:
        log.info(f'{yaml_file}: 0 bad-file entries.')
    except KeyError:
        log.info(f'{yaml_file}: no bad-file field in the ignore block.')

def read_yaml(yaml_file):
    """Reads a yaml file, returns the object

    Parameters
    ==========
    yaml_file: str, input file
    """
    with open(yaml_file) as FILE:
        config = yaml.load(FILE)
    return config

def get_outfile(yaml_file,overwrite=True,extension='fix'):
    """Determines output filename

    Parameters
    ==========
    yaml_file: str, input file 
    overwrite: bool, optional
        write yaml with same name (true), or add extenion (false)
    extension: str, optional
        extention added to output filename if overwrite=False
    """
    if overwrite:
        out_yaml = yaml_file
    else:
        out_yaml = f'{yaml_file}.{extension}'
    return out_yaml

def write_yaml(config_object, yaml_filename):
    """Dumps config object to a yaml file

    Parameters
    ==========
    config_object: `ruamel.yaml.YAML()` object
    yaml_filename: str, output file
    """
    with open(yaml_filename, 'w') as FILE:
        config = yaml.dump(config_object, FILE)


def main():
    """Applies checks, ensures yaml is updated properly"""
    parser = argparse.ArgumentParser(
        description="Automated YAML checker/updater",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "files",
        type=str,
        nargs="+",
        help="YAML files to check/update"
    )
    parser.add_argument(
        "--check",
        "-c",
        action="store_true",
        default=False,
        help="check input yaml",
    )
    parser.add_argument(
        "--overwrite",
        "-o",
        action="store_true",
        default=False,
        help="overwrite input yaml file(s)",
    )
    parser.add_argument(
        "--roundtrip",
        "-rt",
        action="store_true",
        default=False,
        help="read/write input yaml file(s)",
    )
    parser.add_argument(
        "--addnoise",
        action="store_true",
        default=False,
        help="add noise block to input yaml file(s)",
    )
    parser.add_argument(
        "--addoutlier",
        action="store_true",
        default=False,
        help="add outlier block to input yaml file(s)",
    )
    parser.add_argument(
        "--addresults",
        action="store_true",
        default=False,
        help="add results block to input yaml file(s)",
    )
    parser.add_argument(
        "--addcheck",
        action="store_true",
        default=False,
        help="add check block to input yaml file(s)",
    )
    parser.add_argument(
        "--checkcleared",
        action="store_true",
        default=False,
        help="assign cleared: true if check block is empty",
    )
    parser.add_argument(
        "--setvalue",
        nargs=2,
        help="set yaml field to value (2 items)",
    )
    parser.add_argument(
        "--bkv",
        nargs=3,
        help="add block/key/value (3 items) to instantiate new yaml field",
    )
    args = parser.parse_args()

    # maybe we want a "initiate" function to make a new yaml?
    if args.check:
        for ff in args.files:
            log.setLevel('DEBUG')
            fix_toa_info(ff,overwrite=args.overwrite)
            add_niterations(ff,overwrite=args.overwrite)
            add_noise_block(ff,overwrite=args.overwrite)
            add_dmx_block(ff,overwrite=args.overwrite)
            #curate_comments(ff,overwrite=args.overwrite)
            # + outlier block?
            # + check block?
    elif args.roundtrip:
        for ff in args.files:
            config = read_yaml(ff)
            outfile = get_outfile(ff,overwrite=args.overwrite,extension='rt')
            write_yaml(config,outfile)
    elif args.addnoise:
        for ff in args.files:
            add_noise_block(ff,overwrite=args.overwrite)
    elif args.addoutlier:
        for ff in args.files:
            add_outlier_block(ff,overwrite=args.overwrite)
    elif args.addresults:
        for ff in args.files:
            add_results_block(ff,overwrite=args.overwrite)
    elif args.addcheck:
        for ff in args.files:
            add_check_block(ff,overwrite=args.overwrite)
    elif args.checkcleared:
        for ff in args.files:
            check_cleared(ff,overwrite=args.overwrite)
    elif args.setvalue:
        field,value = args.setvalue
        for ff in args.files:
            set_field(ff,field,value,overwrite=args.overwrite)
    elif args.bkv:
        block, key, value = args.bkv
        for ff in args.files:
            add_block_field(ff,block,key,value,overwrite=args.overwrite)

if __name__ == "__main__":
    main()



import pint.toa as toa
import sys

# This is v0 of a script to load in a tim file and bring flags into alignment with the IPTA data standard
# TO-DO
# ------
# -Find edge cases

def reflag_add_empty(toa, flag_name):
    if toa.get_flag_value(flag_name)[0][0] == None:
        toa[flag_name] = 'unknown'
        
def reflag_add_value(toa, flag_name, value):
    if toa.get_flag_value(flag_name)[0][0] == None:
        toa[flag_name] = value
        
def reflag_alt_name(toa, flag_name, alt_name):
    if toa.get_flag_value(flag_name)[0][0] == None:
        print(flag_name, 'not present')
        if toa.get_flag_value(alt_name)[0][0] == None:
            toa[flag_name] = 'unknown'
#            print(alt_name,'not present')
        else:
            toa[flag_name] = toa[alt_name][0] 
#            print(alt_name, 'is present')


def ipta_standard_reflag(toa, f_alt, fe_alt, be_alt, pta):
    # Account for alternative names for fe/be/f
    reflag_alt_name(toa, 'f', f_alt)
    reflag_alt_name(toa, 'fe', fe_alt)
    reflag_alt_name(toa, 'be', be_alt)   
    
    # Add value if flag is missing but the value is known.
    reflag_add_value(toa, 'pta', pta)

    # Add 'unknown' if flags are missing (where there is not a known alternative)
    reflag_add_empty(toa, 'proc')
    reflag_add_empty(toa, 'B')
    reflag_add_empty(toa, 'bw')
    reflag_add_empty(toa, 'tobs')
    #reflag_add_empty(toa, 'fe')
    

    
    
def process_reflagging(timfile, outfile, alt_flag_dict, pta, bipm_ver='BIPM2019', eph='DE440'):
    t = toa.get_TOAs(timfile, bipm_version=bipm_ver,ephem=eph)
    ipta_standard_reflag(t, alt_flag_dict['f'], alt_flag_dict['fe'], alt_flag_dict['be'], pta)
    
    t.write_TOA_file(outfile, format='tempo2')
    
    
ipta_alt_flags = {'f': 'sys', 'fe': 'r', 'be': 'i'}

    
if  __name__=="__main__":
    narg = len(sys.argv)
    timfile = sys.argv[1]
    print('Tim file is', timfile)
    outfile = sys.argv[2]
    print('Outfile name is', outfile)
    pta = sys.argv[3]
    print('PTA is ', pta)
   
    
    process_reflagging(timfile, outfile, ipta_alt_flags, pta)
        
    

    
    

